"""
Numerical/mathematical related functions.

.. versionadded:: 2.1.0
"""

import math
import operator
import typing as t

import pydash as pyd

from .helpers import UNSET, Unset, iterator, iterator_with_default, iteriteratee
from .types import IterateeObjT, NumberNoDecimalT, NumberT, SupportsMul, SupportsRound


if t.TYPE_CHECKING:
    from decimal import Decimal  # pragma: no cover

    from _typeshed import SupportsAdd, SupportsRichComparisonT, SupportsSub  # pragma: no cover


__all__ = (
    "add",
    "ceil",
    "clamp",
    "divide",
    "floor",
    "max_",
    "max_by",
    "mean",
    "mean_by",
    "median",
    "min_",
    "min_by",
    "moving_mean",
    "multiply",
    "power",
    "round_",
    "scale",
    "slope",
    "std_deviation",
    "sum_",
    "sum_by",
    "subtract",
    "transpose",
    "variance",
    "zscore",
)

T = t.TypeVar("T")
T2 = t.TypeVar("T2")
T3 = t.TypeVar("T3")


INFINITY = float("inf")


@t.overload
def add(a: "SupportsAdd[T, T2]", b: T) -> T2:
    ...


@t.overload
def add(a: T, b: "SupportsAdd[T, T2]") -> T2:
    ...


def add(a, b):
    """
    Adds two numbers.

    Args:
        a: First number to add.
        b: Second number to add.

    Returns:
        number

    Example:

        >>> add(10, 5)
        15

    .. versionadded:: 2.1.0

    .. versionchanged:: 3.3.0
        Support adding two numbers when passed as positional arguments.

    .. versionchanged:: 4.0.0
        Only support two argument addition.
    """
    return a + b


@t.overload
def sum_(collection: t.Mapping[t.Any, "SupportsAdd[int, T]"]) -> T:
    ...


@t.overload
def sum_(collection: t.Iterable["SupportsAdd[int, T]"]) -> T:
    ...


def sum_(collection):
    """
    Sum each element in `collection`.

    Args:
        collection: Collection to process or first number to add.

    Returns:
        Result of summation.

    Example:

        >>> sum_([1, 2, 3, 4])
        10

    .. versionadded:: 2.1.0

    .. versionchanged:: 3.3.0
        Support adding two numbers when passed as positional arguments.

    .. versionchanged:: 4.0.0
        Move iteratee support to :func:`sum_by`. Move two argument addition to
        :func:`add`.
    """
    return sum_by(collection)


@t.overload
def sum_by(
    collection: t.Mapping[T, T2],
    iteratee: t.Callable[[T2, T, t.Dict[T, T2]], "SupportsAdd[int, T3]"],
) -> T3:
    ...


@t.overload
def sum_by(
    collection: t.Mapping[T, T2], iteratee: t.Callable[[T2, T], "SupportsAdd[int, T3]"]
) -> T3:
    ...


@t.overload
def sum_by(
    collection: t.Mapping[t.Any, T2], iteratee: t.Callable[[T2], "SupportsAdd[int, T3]"]
) -> T3:
    ...


@t.overload
def sum_by(
    collection: t.Iterable[T], iteratee: t.Callable[[T, int, t.List[T]], "SupportsAdd[int, T2]"]
) -> T2:
    ...


@t.overload
def sum_by(collection: t.Iterable[T], iteratee: t.Callable[[T, int], "SupportsAdd[int, T2]"]) -> T2:
    ...


@t.overload
def sum_by(collection: t.Iterable[T], iteratee: t.Callable[[T], "SupportsAdd[int, T2]"]) -> T2:
    ...


@t.overload
def sum_by(collection: t.Mapping[t.Any, "SupportsAdd[int, T]"], iteratee: None = None) -> T:
    ...


@t.overload
def sum_by(collection: t.Iterable["SupportsAdd[int, T]"], iteratee: None = None) -> T:
    ...


def sum_by(collection, iteratee=None):
    """
    Sum each element in `collection`. If iteratee is passed, each element of `collection` is passed
    through an iteratee before the summation is computed.

    Args:
        collection: Collection to process or first number to add.
        iteratee: Iteratee applied per iteration or second number to add.

    Returns:
        Result of summation.

    Example:

        >>> sum_by([1, 2, 3, 4], lambda x: x ** 2)
        30

    .. versionadded:: 4.0.0
    """
    return sum(result[0] for result in iteriteratee(collection, iteratee))


@t.overload
def mean(collection: t.Mapping[t.Any, "SupportsAdd[int, t.Any]"]) -> float:
    ...


@t.overload
def mean(collection: t.Iterable["SupportsAdd[int, t.Any]"]) -> float:
    ...


def mean(collection):
    """
    Calculate arithmetic mean of each element in `collection`.

    Args:
        collection: Collection to process.

    Returns:
        Result of mean.

    Example:

        >>> mean([1, 2, 3, 4])
        2.5

    .. versionadded:: 2.1.0

    .. versionchanged:: 4.0.0

        - Removed ``average`` and ``avg`` aliases.
        - Moved iteratee functionality to :func:`mean_by`.
    """
    return mean_by(collection)


@t.overload
def mean_by(
    collection: t.Mapping[T, T2],
    iteratee: t.Callable[[T2, T, t.Dict[T, T2]], "SupportsAdd[int, t.Any]"],
) -> float:
    ...


@t.overload
def mean_by(
    collection: t.Mapping[T, T2], iteratee: t.Callable[[T2, T], "SupportsAdd[int, t.Any]"]
) -> float:
    ...


@t.overload
def mean_by(
    collection: t.Mapping[t.Any, T2], iteratee: t.Callable[[T2], "SupportsAdd[int, t.Any]"]
) -> float:
    ...


@t.overload
def mean_by(
    collection: t.Iterable[T], iteratee: t.Callable[[T, int, t.List[T]], "SupportsAdd[int, t.Any]"]
) -> float:
    ...


@t.overload
def mean_by(
    collection: t.Iterable[T], iteratee: t.Callable[[T, int], "SupportsAdd[int, t.Any]"]
) -> float:
    ...


@t.overload
def mean_by(
    collection: t.Iterable[T], iteratee: t.Callable[[T], "SupportsAdd[int, t.Any]"]
) -> float:
    ...


@t.overload
def mean_by(
    collection: t.Mapping[t.Any, "SupportsAdd[int, t.Any]"], iteratee: None = None
) -> float:
    ...


@t.overload
def mean_by(collection: t.Iterable["SupportsAdd[int, t.Any]"], iteratee: None = None) -> float:
    ...


def mean_by(collection, iteratee=None):
    """
    Calculate arithmetic mean of each element in `collection`. If iteratee is passed, each element
    of `collection` is passed through an iteratee before the mean is computed.

    Args:
        collection: Collection to process.
        iteratee: Iteratee applied per iteration.

    Returns:
        Result of mean.

    Example:

        >>> mean_by([1, 2, 3, 4], lambda x: x ** 2)
        7.5

    .. versionadded:: 4.0.0
    """
    return sum_by(collection, iteratee) / len(collection)


def ceil(x: NumberT, precision: int = 0) -> float:
    """
    Round number up to precision.

    Args:
        x: Number to round up.
        precision: Rounding precision. Defaults to ``0``.

    Returns:
        Number rounded up.

    Example:

        >>> ceil(3.275) == 4.0
        True
        >>> ceil(3.215, 1) == 3.3
        True
        >>> ceil(6.004, 2) == 6.01
        True

    .. versionadded:: 3.3.0
    """
    return rounder(math.ceil, x, precision)


NumT = t.TypeVar("NumT", int, float, "Decimal")
NumT2 = t.TypeVar("NumT2", int, float, "Decimal")
NumT3 = t.TypeVar("NumT3", int, float, "Decimal")


def clamp(x: NumT, lower: NumT2, upper: t.Union[NumT3, None] = None) -> t.Union[NumT, NumT2, NumT3]:
    """
    Clamps number within the inclusive lower and upper bounds.

    Args:
        x: Number to clamp.
        lower: Lower bound.
        upper: Upper bound

    Returns:
        number

    Example:

        >>> clamp(-10, -5, 5)
        -5
        >>> clamp(10, -5, 5)
        5
        >>> clamp(10, 5)
        5
        >>> clamp(-10, 5)
        -10

    .. versionadded:: 4.0.0
    """
    if upper is None:
        upper = lower  # type: ignore
        lower = x  # type: ignore

    if x < lower:
        x = lower  # type: ignore
    elif x > upper:  # type: ignore
        x = upper  # type: ignore

    return x


def divide(dividend: t.Union[NumberT, None], divisor: t.Union[NumberT, None]) -> float:
    """
    Divide two numbers.

    Args:
        dividend: The first number in a division.
        divisor: The second number in a division.

    Returns:
        Returns the quotient.

    Example:

        >>> divide(20, 5)
        4.0
        >>> divide(1.5, 3)
        0.5
        >>> divide(None, None)
        1.0
        >>> divide(5, None)
        5.0

    .. versionadded:: 4.0.0
    """
    return call_math_operator(dividend, divisor, operator.truediv, 1)


def floor(x: NumberT, precision: int = 0) -> float:
    """
    Round number down to precision.

    Args:
        x: Number to round down.
        precision: Rounding precision. Defaults to ``0``.

    Returns:
        Number rounded down.

    Example:

        >>> floor(3.75) == 3.0
        True
        >>> floor(3.215, 1) == 3.2
        True
        >>> floor(0.046, 2) == 0.04
        True

    .. versionadded:: 3.3.0
    """
    return rounder(math.floor, x, precision)


@t.overload
def max_(
    collection: t.Mapping[t.Any, "SupportsRichComparisonT"], default: Unset = UNSET
) -> "SupportsRichComparisonT":
    ...


@t.overload
def max_(
    collection: t.Mapping[t.Any, "SupportsRichComparisonT"], default: T
) -> t.Union["SupportsRichComparisonT", T]:
    ...


@t.overload
def max_(
    collection: t.Iterable["SupportsRichComparisonT"], default: Unset = UNSET
) -> "SupportsRichComparisonT":
    ...


@t.overload
def max_(
    collection: t.Iterable["SupportsRichComparisonT"], default: T
) -> t.Union["SupportsRichComparisonT", T]:
    ...


def max_(collection, default=UNSET):
    """
    Retrieves the maximum value of a `collection`.

    Args:
        collection: Collection to iterate over.
        default: Value to return if `collection` is empty.

    Returns:
        Maximum value.

    Example:

        >>> max_([1, 2, 3, 4])
        4
        >>> max_([], default=-1)
        -1

    .. versionadded:: 1.0.0

    .. versionchanged:: 4.0.0
        Moved iteratee iteratee support to :func:`max_by`.
    """
    return max_by(collection, default=default)


@t.overload
def max_by(
    collection: t.Mapping[t.Any, "SupportsRichComparisonT"],
    iteratee: None = None,
    default: Unset = UNSET,
) -> "SupportsRichComparisonT":
    ...


@t.overload
def max_by(
    collection: t.Mapping[t.Any, T2],
    iteratee: t.Callable[[T2], "SupportsRichComparisonT"],
    default: Unset = UNSET,
) -> T2:
    ...


@t.overload
def max_by(
    collection: t.Mapping[t.Any, T2],
    iteratee: t.Callable[[T2], "SupportsRichComparisonT"],
    *,
    default: T
) -> t.Union[T2, T]:
    ...


@t.overload
def max_by(
    collection: t.Mapping[t.Any, "SupportsRichComparisonT"], iteratee: None = None, *, default: T
) -> t.Union["SupportsRichComparisonT", T]:
    ...


@t.overload
def max_by(
    collection: t.Iterable["SupportsRichComparisonT"], iteratee: None = None, default: Unset = UNSET
) -> "SupportsRichComparisonT":
    ...


@t.overload
def max_by(
    collection: t.Iterable[T2],
    iteratee: t.Callable[[T2], "SupportsRichComparisonT"],
    default: Unset = UNSET,
) -> T2:
    ...


@t.overload
def max_by(
    collection: t.Iterable[T2], iteratee: t.Callable[[T2], "SupportsRichComparisonT"], *, default: T
) -> t.Union[T2, T]:
    ...


@t.overload
def max_by(
    collection: t.Iterable["SupportsRichComparisonT"], iteratee: None = None, *, default: T
) -> t.Union["SupportsRichComparisonT", T]:
    ...


@t.overload
def max_by(collection: t.Iterable[T], iteratee: IterateeObjT, default: Unset = UNSET) -> T:
    ...


@t.overload
def max_by(collection: t.Iterable[T], iteratee: IterateeObjT, default: T2) -> t.Union[T, T2]:
    ...


def max_by(collection, iteratee=None, default=UNSET):
    """
    Retrieves the maximum value of a `collection`.

    Args:
        collection: Collection to iterate over.
        iteratee: Iteratee applied per iteration.
        default: Value to return if `collection` is empty.

    Returns:
        Maximum value.

    Example:

        >>> max_by([1.0, 1.5, 1.8], math.floor)
        1.0
        >>> max_by([{'a': 1}, {'a': 2}, {'a': 3}], 'a')
        {'a': 3}
        >>> max_by([], default=-1)
        -1

    .. versionadded:: 4.0.0
    """
    if isinstance(collection, dict):
        collection = collection.values()

    return max(iterator_with_default(collection, default), key=pyd.iteratee(iteratee))


@t.overload
def median(
    collection: t.Mapping[T, T2], iteratee: t.Callable[[T2, T, t.Dict[T, T2]], NumberT]
) -> t.Union[float, int]:
    ...


@t.overload
def median(
    collection: t.Mapping[T, T2], iteratee: t.Callable[[T2, T], NumberT]
) -> t.Union[float, int]:
    ...


@t.overload
def median(
    collection: t.Mapping[t.Any, T2], iteratee: t.Callable[[T2], NumberT]
) -> t.Union[float, int]:
    ...


@t.overload
def median(
    collection: t.Iterable[T], iteratee: t.Callable[[T, int, t.List[T]], NumberT]
) -> t.Union[float, int]:
    ...


@t.overload
def median(
    collection: t.Iterable[T], iteratee: t.Callable[[T, int], NumberT]
) -> t.Union[float, int]:
    ...


@t.overload
def median(collection: t.Iterable[T], iteratee: t.Callable[[T], NumberT]) -> t.Union[float, int]:
    ...


@t.overload
def median(collection: t.Iterable[NumberT], iteratee: None = None) -> t.Union[float, int]:
    ...


def median(collection, iteratee=None):
    """
    Calculate median of each element in `collection`. If iteratee is passed, each element of
    `collection` is passed through an iteratee before the median is computed.

    Args:
        collection: Collection to process.
        iteratee: Iteratee applied per iteration.

    Returns:
        Result of median.

    Example:

        >>> median([1, 2, 3, 4, 5])
        3
        >>> median([1, 2, 3, 4])
        2.5

    .. versionadded:: 2.1.0
    """
    length = len(collection)
    middle = (length + 1) / 2
    collection = sorted(ret[0] for ret in iteriteratee(collection, iteratee))

    if pyd.is_odd(length):
        result = collection[int(middle - 1)]
    else:
        left = int(middle - 1.5)
        right = int(middle - 0.5)
        result = (collection[left] + collection[right]) / 2

    return result


@t.overload
def min_(
    collection: t.Mapping[t.Any, "SupportsRichComparisonT"], default: Unset = UNSET
) -> "SupportsRichComparisonT":
    ...


@t.overload
def min_(
    collection: t.Mapping[t.Any, "SupportsRichComparisonT"], default: T
) -> t.Union["SupportsRichComparisonT", T]:
    ...


@t.overload
def min_(
    collection: t.Iterable["SupportsRichComparisonT"], default: Unset = UNSET
) -> "SupportsRichComparisonT":
    ...


@t.overload
def min_(
    collection: t.Iterable["SupportsRichComparisonT"], default: T
) -> t.Union["SupportsRichComparisonT", T]:
    ...


def min_(collection, default=UNSET):
    """
    Retrieves the minimum value of a `collection`.

    Args:
        collection: Collection to iterate over.
        default: Value to return if `collection` is empty.

    Returns:
        Minimum value.

    Example:

        >>> min_([1, 2, 3, 4])
        1
        >>> min_([], default=100)
        100

    .. versionadded:: 1.0.0

    .. versionchanged:: 4.0.0
        Moved iteratee iteratee support to :func:`min_by`.
    """
    return min_by(collection, default=default)


@t.overload
def min_by(
    collection: t.Mapping[t.Any, "SupportsRichComparisonT"],
    iteratee: None = None,
    default: Unset = UNSET,
) -> "SupportsRichComparisonT":
    ...


@t.overload
def min_by(
    collection: t.Mapping[t.Any, T2],
    iteratee: t.Callable[[T2], "SupportsRichComparisonT"],
    default: Unset = UNSET,
) -> T2:
    ...


@t.overload
def min_by(
    collection: t.Mapping[t.Any, T2],
    iteratee: t.Callable[[T2], "SupportsRichComparisonT"],
    *,
    default: T
) -> t.Union[T2, T]:
    ...


@t.overload
def min_by(
    collection: t.Mapping[t.Any, "SupportsRichComparisonT"], iteratee: None = None, *, default: T
) -> t.Union["SupportsRichComparisonT", T]:
    ...


@t.overload
def min_by(
    collection: t.Iterable["SupportsRichComparisonT"], iteratee: None = None, default: Unset = UNSET
) -> "SupportsRichComparisonT":
    ...


@t.overload
def min_by(
    collection: t.Iterable[T2],
    iteratee: t.Callable[[T2], "SupportsRichComparisonT"],
    default: Unset = UNSET,
) -> T2:
    ...


@t.overload
def min_by(
    collection: t.Iterable[T2], iteratee: t.Callable[[T2], "SupportsRichComparisonT"], *, default: T
) -> t.Union[T2, T]:
    ...


@t.overload
def min_by(
    collection: t.Iterable["SupportsRichComparisonT"], iteratee: None = None, *, default: T
) -> t.Union["SupportsRichComparisonT", T]:
    ...


@t.overload
def min_by(collection: t.Iterable[T], iteratee: IterateeObjT, default: Unset = UNSET) -> T:
    ...


@t.overload
def min_by(collection: t.Iterable[T], iteratee: IterateeObjT, default: T2) -> t.Union[T, T2]:
    ...


def min_by(collection, iteratee=None, default=UNSET):
    """
    Retrieves the minimum value of a `collection`.

    Args:
        collection: Collection to iterate over.
        iteratee: Iteratee applied per iteration.
        default: Value to return if `collection` is empty.

    Returns:
        Minimum value.

    Example:

        >>> min_by([1.8, 1.5, 1.0], math.floor)
        1.8
        >>> min_by([{'a': 1}, {'a': 2}, {'a': 3}], 'a')
        {'a': 1}
        >>> min_by([], default=100)
        100

    .. versionadded:: 4.0.0
    """
    if isinstance(collection, dict):
        collection = collection.values()
    return min(iterator_with_default(collection, default), key=pyd.iteratee(iteratee))


def moving_mean(array: t.Sequence["SupportsAdd[int, t.Any]"], size: t.SupportsInt) -> t.List[float]:
    """
    Calculate moving mean of each element of `array`.

    Args:
        array: List to process.
        size: Window size.

    Returns:
        Result of moving average.

    Example:

        >>> moving_mean(range(10), 1)
        [0.0, 1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 7.0, 8.0, 9.0]
        >>> moving_mean(range(10), 5)
        [2.0, 3.0, 4.0, 5.0, 6.0, 7.0]
        >>> moving_mean(range(10), 10)
        [4.5]

    .. versionadded:: 2.1.0

    .. versionchanged:: 4.0.0
        Rename to ``moving_mean`` and remove ``moving_average`` and ``moving_avg`` aliases.
    """
    result = []
    size = int(size)

    for i in range(size - 1, len(array) + 1):
        window = array[i - size : i]

        if len(window) == size:
            result.append(mean(window))

    return result


@t.overload
def multiply(multiplier: SupportsMul[int, T2], multiplicand: None) -> T2:
    ...


@t.overload
def multiply(multiplier: None, multiplicand: SupportsMul[int, T2]) -> T2:
    ...


@t.overload
def multiply(multiplier: None, multiplicand: None) -> int:
    ...


@t.overload
def multiply(multiplier: SupportsMul[T, T2], multiplicand: T) -> T2:
    ...


@t.overload
def multiply(multiplier: T, multiplicand: SupportsMul[T, T2]) -> T2:
    ...


def multiply(multiplier, multiplicand):
    """
    Multiply two numbers.

    Args:
        multiplier: The first number in a multiplication.
        multiplicand: The second number in a multiplication.

    Returns:
        Returns the product.

    Example:

        >>> multiply(4, 5)
        20
        >>> multiply(10, 4)
        40
        >>> multiply(None, 10)
        10
        >>> multiply(None, None)
        1

    .. versionadded:: 4.0.0
    """
    return call_math_operator(multiplier, multiplicand, operator.mul, 1)


@t.overload
def power(x: int, n: int) -> t.Union[int, float]:
    ...


@t.overload
def power(x: float, n: t.Union[int, float]) -> float:
    ...


@t.overload
def power(x: t.List[int], n: int) -> t.List[t.Union[int, float]]:
    ...


@t.overload
def power(x: t.List[float], n: t.List[t.Union[int, float]]) -> t.List[float]:
    ...


def power(x, n):
    """
    Calculate exponentiation of `x` raised to the `n` power.

    Args:
        x: Base number.
        n: Exponent.

    Returns:
        Result of calculation.

    Example:

        >>> power(5, 2)
        25
        >>> power(12.5, 3)
        1953.125

    .. versionadded:: 2.1.0

    .. versionchanged:: 4.0.0
        Removed alias ``pow_``.
    """
    if pyd.is_number(x):
        result = pow(x, n)
    elif pyd.is_list(x):
        result = [pow(item, n) for item in x]
    else:
        result = None

    return result


@t.overload
def round_(x: t.List[SupportsRound[NumberT]], precision: int = 0) -> t.List[float]:
    ...


@t.overload
def round_(x: SupportsRound[NumberT], precision: int = 0) -> float:
    ...


def round_(x, precision=0):
    """
    Round number to precision.

    Args:
        x: Number to round.
        precision: Rounding precision. Defaults to ``0``.

    Returns:
        Rounded number.

    Example:

        >>> round_(3.275) == 3.0
        True
        >>> round_(3.275, 1) == 3.3
        True

    .. versionadded:: 2.1.0

    .. versionchanged:: 4.0.0
        Remove alias ``curve``.
    """
    return rounder(round, x, precision)


@t.overload
def scale(array: t.Iterable["Decimal"], maximum: "Decimal") -> t.List["Decimal"]:
    ...


@t.overload
def scale(array: t.Iterable[NumberNoDecimalT], maximum: NumberNoDecimalT) -> t.List[float]:
    ...


@t.overload
def scale(array: t.Iterable[NumberT], maximum: int = 1) -> t.List[float]:
    ...


def scale(array, maximum: NumberT = 1):
    """
    Scale list of value to a maximum number.

    Args:
        array: Numbers to scale.
        maximum: Maximum scale value.

    Returns:
        Scaled numbers.

    Example:

        >>> scale([1, 2, 3, 4])
        [0.25, 0.5, 0.75, 1.0]
        >>> scale([1, 2, 3, 4], 1)
        [0.25, 0.5, 0.75, 1.0]
        >>> scale([1, 2, 3, 4], 4)
        [1.0, 2.0, 3.0, 4.0]
        >>> scale([1, 2, 3, 4], 2)
        [0.5, 1.0, 1.5, 2.0]

    .. versionadded:: 2.1.0
    """
    array_max = max(array)
    factor = maximum / array_max
    return [item * factor for item in array]


@t.overload
def slope(
    point1: t.Union[t.Tuple["Decimal", "Decimal"], t.List["Decimal"]],
    point2: t.Union[t.Tuple["Decimal", "Decimal"], t.List["Decimal"]],
) -> "Decimal":
    ...


@t.overload
def slope(
    point1: t.Union[t.Tuple[NumberNoDecimalT, NumberNoDecimalT], t.List[NumberNoDecimalT]],
    point2: t.Union[t.Tuple[NumberNoDecimalT, NumberNoDecimalT], t.List[NumberNoDecimalT]],
) -> float:
    ...


def slope(point1, point2):
    """
    Calculate the slope between two points.

    Args:
        point1: X and Y coordinates of first point.
        point2: X and Y cooredinates of second point.

    Returns:
        Calculated slope.

    Example:

        >>> slope((1, 2), (4, 8))
        2.0

    .. versionadded:: 2.1.0
    """
    x1, y1 = point1[0], point1[1]
    x2, y2 = point2[0], point2[1]

    if x1 == x2:
        result = INFINITY
    else:
        result = (y2 - y1) / (x2 - x1)

    return result


def std_deviation(array: t.List[NumberT]) -> float:
    """
    Calculate standard deviation of list of numbers.

    Args:
        array: List to process.

    Returns:
        Calculated standard deviation.

    Example:

        >>> round(std_deviation([1, 18, 20, 4]), 2) == 8.35
        True

    .. versionadded:: 2.1.0

    .. versionchanged:: 4.0.0
        Remove alias ``sigma``.
    """
    return math.sqrt(variance(array))


@t.overload
def subtract(minuend: "SupportsSub[T, T2]", subtrahend: T) -> T2:
    ...


@t.overload
def subtract(minuend: T, subtrahend: "SupportsSub[T, T2]") -> T2:
    ...


def subtract(minuend, subtrahend):
    """
    Subtracts two numbers.

    Args:
        minuend: Value passed in by the user.
        subtrahend: Value passed in by the user.

    Returns:
        Result of the difference from the given values.

    Example:

        >>> subtract(10, 5)
        5
        >>> subtract(-10, 4)
        -14
        >>> subtract(2, 0.5)
        1.5

    .. versionadded:: 4.0.0
    """
    return call_math_operator(minuend, subtrahend, operator.sub, 0)


def transpose(array: t.Iterable[t.Iterable[T]]) -> t.List[t.List[T]]:
    """
    Transpose the elements of `array`.

    Args:
        array: List to process.

    Returns:
        Transposed list.

    Example:

        >>> transpose([[1, 2, 3], [4, 5, 6], [7, 8, 9]])
        [[1, 4, 7], [2, 5, 8], [3, 6, 9]]

    .. versionadded:: 2.1.0
    """
    trans: t.List[t.List[T]] = []

    for y, row in iterator(array):
        for x, col in iterator(row):
            trans = pyd.set_(trans, [x, y], col)

    return trans


@t.overload
def variance(array: t.Mapping[t.Any, "SupportsAdd[int, t.Any]"]) -> float:
    ...


@t.overload
def variance(array: t.Iterable["SupportsAdd[int, t.Any]"]) -> float:
    ...


def variance(array):
    """
    Calculate the variance of the elements in `array`.

    Args:
        array: List to process.

    Returns:
        Calculated variance.

    Example:

        >>> variance([1, 18, 20, 4])
        69.6875

    .. versionadded:: 2.1.0
    """
    avg = mean(array)

    def var(x):
        return power(x - avg, 2)

    return pyd._(array).map_(var).mean().value()


@t.overload
def zscore(
    collection: t.Mapping[T, T2], iteratee: t.Callable[[T2, T, t.Dict[T, T2]], NumberT]
) -> t.List[float]:
    ...


@t.overload
def zscore(collection: t.Mapping[T, T2], iteratee: t.Callable[[T2, T], NumberT]) -> t.List[float]:
    ...


@t.overload
def zscore(collection: t.Mapping[t.Any, T2], iteratee: t.Callable[[T2], NumberT]) -> t.List[float]:
    ...


@t.overload
def zscore(
    collection: t.Iterable[T], iteratee: t.Callable[[T, int, t.List[T]], NumberT]
) -> t.List[float]:
    ...


@t.overload
def zscore(collection: t.Iterable[T], iteratee: t.Callable[[T, int], NumberT]) -> t.List[float]:
    ...


@t.overload
def zscore(collection: t.Iterable[T], iteratee: t.Callable[[T], NumberT]) -> t.List[float]:
    ...


@t.overload
def zscore(collection: t.Iterable[NumberT], iteratee: None = None) -> t.List[float]:
    ...


def zscore(collection, iteratee=None):
    """
    Calculate the standard score assuming normal distribution. If iteratee is passed, each element
    of `collection` is passed through an iteratee before the standard score is computed.

    Args:
        collection: Collection to process.
        iteratee: Iteratee applied per iteration.

    Returns:
        Calculated standard score.

    Example:

        >>> results = zscore([1, 2, 3])

        # [-1.224744871391589, 0.0, 1.224744871391589]

    .. versionadded:: 2.1.0
    """
    array = pyd.map_(collection, iteratee)
    avg = mean(array)
    sig = std_deviation(array)

    return [(item - avg) / sig for item in array]


#
# Utility methods not a part of the main API
#


def call_math_operator(value1, value2, op, default):
    """Return the result of the math operation on the given values."""
    if not value1:
        value1 = default

    if not value2:
        value2 = default

    if not pyd.is_number(value1):
        try:
            value1 = float(value1)
        except Exception:
            pass

    if not pyd.is_number(value2):
        try:
            value2 = float(value2)
        except Exception:
            pass

    return op(value1, value2)


def rounder(func, x, precision):
    precision = pow(10, precision)

    def rounder_func(item):
        return func(item * precision) / precision

    result = None

    if pyd.is_number(x):
        result = rounder_func(x)
    elif pyd.is_iterable(x):
        try:
            result = [rounder_func(item) for item in x]
        except TypeError:
            pass

    return result



import math
import typing as t

import pytest

import pydash as _


@pytest.mark.mypy_testing
def test_mypy_add() -> None:
    reveal_type(_.add(10, 5))  # R: builtins.int
    reveal_type(_.add(10.1, 5))  # R: builtins.float
    reveal_type(_.add(10, 5.5))  # R: builtins.float


@pytest.mark.mypy_testing
def test_mypy_sum_() -> None:
    reveal_type(_.sum_([1, 2, 3, 4]))  # R: builtins.int
    reveal_type(_.sum_([1.5, 2, 3, 4]))  # R: builtins.float
    reveal_type(_.sum_({"hello": 1, "bye": 2}))  # R: builtins.int


@pytest.mark.mypy_testing
def test_mypy_sum_by() -> None:
    reveal_type(_.sum_by([1, 2, 3, 4], lambda x: x ** 2))  # R: builtins.int


@pytest.mark.mypy_testing
def test_mypy_mean() -> None:
    reveal_type(_.mean([1, 2, 3, 4]))  # R: builtins.float


@pytest.mark.mypy_testing
def test_mypy_mean_by() -> None:
    reveal_type(_.mean_by([1, 2, 3, 4], lambda x: x ** 2))  # R: builtins.float


@pytest.mark.mypy_testing
def test_mypy_ceil() -> None:
    reveal_type(_.ceil(3.275))  # R: builtins.float
    reveal_type(_.ceil(3.215, 1))  # R: builtins.float
    reveal_type(_.ceil(6.004, 2))  # R: builtins.float


@pytest.mark.mypy_testing
def test_mypy_clamp() -> None:
    reveal_type(_.clamp(-10, -5, 5))  # R: builtins.int
    reveal_type(_.clamp(10, -5, 5))  # R: builtins.int
    reveal_type(_.clamp(10, 5))  # R: builtins.int
    reveal_type(_.clamp(-10, 5))  # R: builtins.int


@pytest.mark.mypy_testing
def test_mypy_divide() -> None:
    reveal_type(_.divide(20, 5))  # R: builtins.float
    reveal_type(_.divide(1.5, 3))  # R: builtins.float
    reveal_type(_.divide(None, None))  # R: builtins.float
    reveal_type(_.divide(5, None))  # R: builtins.float


@pytest.mark.mypy_testing
def test_mypy_floor() -> None:
    reveal_type(_.floor(3.75))  # R: builtins.float
    reveal_type(_.floor(3.215, 1))  # R: builtins.float
    reveal_type(_.floor(0.046, 2))  # R: builtins.float


@pytest.mark.mypy_testing
def test_mypy_max_() -> None:
    reveal_type(_.max_([1, 2, 3, 4]))  # R: builtins.int

    empty_int_list: t.List[int] = []
    reveal_type(_.max_(empty_int_list, default=-1))  # R: builtins.int


@pytest.mark.mypy_testing
def test_mypy_max_by() -> None:
    def floor(x: float) -> int:
        return math.floor(x)

    reveal_type(_.max_by([1.0, 1.5, 1.8], floor))  # R: builtins.float
    reveal_type(_.max_by([{'a': 1}, {'a': 2}, {'a': 3}], 'a'))  # R: builtins.dict[builtins.str, builtins.int]

    empty_int_list: t.List[int] = []
    reveal_type(_.max_by(empty_int_list, default=-1))  # R: builtins.int


@pytest.mark.mypy_testing
def test_mypy_median() -> None:
    reveal_type(_.median([1, 2, 3, 4, 5]))  # R: Union[builtins.float, builtins.int]
    reveal_type(_.median([1, 2, 3, 4]))  # R: Union[builtins.float, builtins.int]


@pytest.mark.mypy_testing
def test_mypy_min_() -> None:
    reveal_type(_.min_([1, 2, 3, 4]))  # R: builtins.int

    empty_int_list: t.List[int] = []
    reveal_type(_.min_(empty_int_list, default=100))  # R: builtins.int


@pytest.mark.mypy_testing
def test_mypy_min_by() -> None:
    def floor(x: float) -> int:
        return math.floor(x)

    reveal_type(_.min_by([1.8, 1.5, 1.0], floor))  # R: builtins.float
    reveal_type(_.min_by([{'a': 1}, {'a': 2}, {'a': 3}], 'a'))  # R: builtins.dict[builtins.str, builtins.int]

    empty_int_list: t.List[int] = []
    reveal_type(_.min_by(empty_int_list, default=100))  # R: builtins.int


@pytest.mark.mypy_testing
def test_mypy_moving_mean() -> None:
    reveal_type(_.moving_mean(range(10), 1))  # R: builtins.list[builtins.float]


@pytest.mark.mypy_testing
def test_mypy_multiply() -> None:
    reveal_type(_.multiply(4, 5))  # R: builtins.int
    reveal_type(_.multiply(10.5, 4))  # R: builtins.float
    reveal_type(_.multiply(10, 4.5))  # R: builtins.float
    reveal_type(_.multiply(None, 10))  # R: builtins.int
    reveal_type(_.multiply(10, None))  # R: builtins.int
    reveal_type(_.multiply(None, 5.5))  # R: builtins.float
    reveal_type(_.multiply(5.5, None))  # R: builtins.float
    reveal_type(_.multiply(None, None))  # R: builtins.int


@pytest.mark.mypy_testing
def test_mypy_power() -> None:
    reveal_type(_.power(5, 2))  # R: Union[builtins.int, builtins.float]
    reveal_type(_.power(12.5, 3))  # R: builtins.float


@pytest.mark.mypy_testing
def test_mypy_round_() -> None:
    reveal_type(_.round_(3.275))  # R: builtins.float
    reveal_type(_.round_(3.275, 1))  # R: builtins.float


@pytest.mark.mypy_testing
def test_mypy_scale() -> None:
    reveal_type(_.scale([1, 2, 3, 4]))  # R: builtins.list[builtins.float]
    reveal_type(_.scale([1, 2, 3, 4], 1))  # R: builtins.list[builtins.float]


@pytest.mark.mypy_testing
def test_mypy_slope() -> None:
    reveal_type(_.slope((1, 2), (4, 8)))  # R: builtins.float


@pytest.mark.mypy_testing
def test_mypy_std_deviation() -> None:
    reveal_type(_.std_deviation([1, 18, 20, 4]))  # R: builtins.float


@pytest.mark.mypy_testing
def test_mypy_subtract() -> None:
    reveal_type(_.subtract(10, 5))  # R: builtins.int
    reveal_type(_.subtract(-10, 4))  # R: builtins.int
    reveal_type(_.subtract(2, 0.5))  # R: builtins.float


@pytest.mark.mypy_testing
def test_mypy_transpose() -> None:
    reveal_type(_.transpose([[1, 2, 3], [4, 5, 6], [7, 8, 9]]))  # R: builtins.list[builtins.list[builtins.int]]


@pytest.mark.mypy_testing
def test_mypy_variance() -> None:
    reveal_type(_.variance([1, 18, 20, 4]))  # R: builtins.float


@pytest.mark.mypy_testing
def test_mypy_var() -> None:
    reveal_type(_.zscore([1, 2, 3]))  # R: builtins.list[builtins.float]



from abc import ABC, abstractmethod
import typing as t


class AllFuncs(ABC):
    """Exposing all of the exposed functions of a module through an class."""

    module: t.Any
    invalid_method_exception: t.Type[Exception]

    @abstractmethod
    def _wrap(self, func) -> t.Callable:
        """Proxy attribute access to :attr:`module`."""
        raise NotImplementedError()  # pragma: no cover

    @classmethod
    def get_method(cls, name: str) -> t.Callable:
        """
        Return valid :attr:`module` method.

        Args:
            name: Name of pydash method to get.

        Returns:
            :attr:`module` callable.

        Raises:
            InvalidMethod: Raised if `name` is not a valid :attr:`module` method.
        """
        method = getattr(cls.module, name, None)

        if not callable(method) and not name.endswith("_"):
            # Alias method names not ending in underscore to their underscore
            # counterpart. This allows chaining of functions like "map_()"
            # using "map()" instead.
            method = getattr(cls.module, name + "_", None)

        if not callable(method):
            raise cls.invalid_method_exception(f"Invalid {cls.module.__name__} method: {name}")

        return method

    def __getattr__(self, name: str) -> t.Callable:
        return self._wrap(self.get_method(name))



"""
Functions that operate on lists and dicts.

.. versionadded:: 1.0.0
"""

from functools import cmp_to_key
import random
import typing as t

import pydash as pyd

from .helpers import callit, cmp, getargcount, iterator, iteriteratee
from .types import IterateeObjT, PathT


__all__ = (
    "at",
    "count_by",
    "every",
    "filter_",
    "find",
    "find_last",
    "flat_map",
    "flat_map_deep",
    "flat_map_depth",
    "for_each",
    "for_each_right",
    "group_by",
    "includes",
    "invoke_map",
    "key_by",
    "map_",
    "nest",
    "order_by",
    "partition",
    "pluck",
    "reduce_",
    "reduce_right",
    "reductions",
    "reductions_right",
    "reject",
    "sample",
    "sample_size",
    "shuffle",
    "size",
    "some",
    "sort_by",
)

T = t.TypeVar("T")
T2 = t.TypeVar("T2")
T3 = t.TypeVar("T3")
T4 = t.TypeVar("T4")


@t.overload
def at(collection: t.Mapping[T, T2], *paths: T) -> t.List[t.Union[T2, None]]:
    ...


@t.overload
def at(collection: t.Mapping[T, t.Any], *paths: t.Union[T, t.Iterable[T]]) -> t.List[t.Any]:
    ...


@t.overload
def at(collection: t.Iterable[T], *paths: int) -> t.List[t.Union[T, None]]:
    ...


@t.overload
def at(collection: t.Iterable[t.Any], *paths: t.Union[int, t.Iterable[int]]) -> t.List[t.Any]:
    ...


def at(collection, *paths):
    """
    Creates a list of elements from the specified indexes, or keys, of the collection. Indexes may
    be specified as individual arguments or as arrays of indexes.

    Args:
        collection: Collection to iterate over.
        *paths: The indexes of `collection` to retrieve, specified as individual indexes or
            arrays of indexes.

    Returns:
        filtered list

    Example:

        >>> at([1, 2, 3, 4], 0, 2)
        [1, 3]
        >>> at({'a': 1, 'b': 2, 'c': 3, 'd': 4}, 'a', 'c')
        [1, 3]
        >>> at({'a': 1, 'b': 2, 'c': {'d': {'e': 3}}}, 'a', ['c', 'd', 'e'])
        [1, 3]

    .. versionadded:: 1.0.0

    .. versionchanged:: 4.1.0
        Support deep path access.
    """
    return pyd.properties(*paths)(collection)


@t.overload
def count_by(collection: t.Mapping[t.Any, T2], iteratee: None = None) -> t.Dict[T2, int]:
    ...


@t.overload
def count_by(
    collection: t.Mapping[T, T2], iteratee: t.Callable[[T2, T, t.Dict[T, T2]], T3]
) -> t.Dict[T3, int]:
    ...


@t.overload
def count_by(collection: t.Mapping[T, T2], iteratee: t.Callable[[T2, T], T3]) -> t.Dict[T3, int]:
    ...


@t.overload
def count_by(collection: t.Mapping[t.Any, T2], iteratee: t.Callable[[T2], T3]) -> t.Dict[T3, int]:
    ...


@t.overload
def count_by(collection: t.Iterable[T], iteratee: None = None) -> t.Dict[T, int]:
    ...


@t.overload
def count_by(
    collection: t.Iterable[T], iteratee: t.Callable[[T, int, t.List[T]], T2]
) -> t.Dict[T2, int]:
    ...


@t.overload
def count_by(collection: t.Iterable[T], iteratee: t.Callable[[T, int], T2]) -> t.Dict[T2, int]:
    ...


@t.overload
def count_by(collection: t.Iterable[T], iteratee: t.Callable[[T], T2]) -> t.Dict[T2, int]:
    ...


def count_by(collection, iteratee=None):
    """
    Creates an object composed of keys generated from the results of running each element of
    `collection` through the iteratee.

    Args:
        collection: Collection to iterate over.
        iteratee: Iteratee applied per iteration.

    Returns:
        Dict containing counts by key.

    Example:

        >>> results = count_by([1, 2, 1, 2, 3, 4])
        >>> assert results == {1: 2, 2: 2, 3: 1, 4: 1}
        >>> results = count_by(['a', 'A', 'B', 'b'], lambda x: x.lower())
        >>> assert results == {'a': 2, 'b': 2}
        >>> results = count_by({'a': 1, 'b': 1, 'c': 3, 'd': 3})
        >>> assert results == {1: 2, 3: 2}

    .. versionadded:: 1.0.0
    """
    ret = {}

    for result in iteriteratee(collection, iteratee):
        ret.setdefault(result[0], 0)
        ret[result[0]] += 1

    return ret


def every(
    collection: t.Iterable[T], predicate: t.Union[t.Callable[[T], t.Any], IterateeObjT, None] = None
) -> bool:
    """
    Checks if the predicate returns a truthy value for all elements of a collection. The predicate
    is invoked with three arguments: ``(value, index|key, collection)``. If a property name is
    passed for predicate, the created :func:`pluck` style predicate will return the property value
    of the given element. If an object is passed for predicate, the created :func:`.matches` style
    predicate will return ``True`` for elements that have the properties of the given object, else
    ``False``.

    Args:
        collection: Collection to iterate over.
        predicate: Predicate applied per iteration.

    Returns:
        Whether all elements are truthy.

    Example:

        >>> every([1, True, 'hello'])
        True
        >>> every([1, False, 'hello'])
        False
        >>> every([{'a': 1}, {'a': True}, {'a': 'hello'}], 'a')
        True
        >>> every([{'a': 1}, {'a': False}, {'a': 'hello'}], 'a')
        False
        >>> every([{'a': 1}, {'a': 1}], {'a': 1})
        True
        >>> every([{'a': 1}, {'a': 2}], {'a': 1})
        False

    .. versionadded:: 1.0.0

    .. versionchanged: 4.0.0
        Removed alias ``all_``.
    """
    if predicate:
        cbk = pyd.iteratee(predicate)
        collection = (cbk(item) for item in collection)

    return all(collection)


@t.overload
def filter_(
    collection: t.Mapping[T, T2],
    predicate: t.Union[t.Callable[[T2, T, t.Dict[T, T2]], t.Any], IterateeObjT, None] = None,
) -> t.List[T2]:
    ...


@t.overload
def filter_(
    collection: t.Mapping[T, T2],
    predicate: t.Union[t.Callable[[T2, T], t.Any], IterateeObjT, None] = None,
) -> t.List[T2]:
    ...


@t.overload
def filter_(
    collection: t.Mapping[t.Any, T2],
    predicate: t.Union[t.Callable[[T2], t.Any], IterateeObjT, None] = None,
) -> t.List[T2]:
    ...


@t.overload
def filter_(
    collection: t.Iterable[T],
    predicate: t.Union[t.Callable[[T, int, t.List[T]], t.Any], IterateeObjT, None] = None,
) -> t.List[T]:
    ...


@t.overload
def filter_(
    collection: t.Iterable[T],
    predicate: t.Union[t.Callable[[T, int], t.Any], IterateeObjT, None] = None,
) -> t.List[T]:
    ...


@t.overload
def filter_(
    collection: t.Iterable[T],
    predicate: t.Union[t.Callable[[T], t.Any], IterateeObjT, None] = None,
) -> t.List[T]:
    ...


def filter_(collection, predicate=None):
    """
    Iterates over elements of a collection, returning a list of all elements the predicate returns
    truthy for.

    Args:
        collection: Collection to iterate over.
        predicate: Predicate applied per iteration.

    Returns:
        Filtered list.

    Example:

        >>> results = filter_([{'a': 1}, {'b': 2}, {'a': 1, 'b': 3}], {'a': 1})
        >>> assert results == [{'a': 1}, {'a': 1, 'b': 3}]
        >>> filter_([1, 2, 3, 4], lambda x: x >= 3)
        [3, 4]

    .. versionadded:: 1.0.0

    .. versionchanged:: 4.0.0
        Removed alias ``select``.
    """
    return [value for is_true, value, _, _ in iteriteratee(collection, predicate) if is_true]


@t.overload
def find(
    collection: t.Dict[T, T2],
    predicate: t.Union[t.Callable[[T2, T, t.Dict[T, T2]], t.Any], IterateeObjT, None] = None,
) -> t.Union[T2, None]:
    ...


@t.overload
def find(
    collection: t.Dict[T, T2],
    predicate: t.Union[t.Callable[[T2, T], t.Any], IterateeObjT, None] = None,
) -> t.Union[T2, None]:
    ...


@t.overload
def find(
    collection: t.Dict[T, T2],
    predicate: t.Union[t.Callable[[T2], t.Any], IterateeObjT, None] = None,
) -> t.Union[T2, None]:
    ...


@t.overload
def find(
    collection: t.List[T],
    predicate: t.Union[t.Callable[[T, int, t.List[T]], t.Any], IterateeObjT, None] = None,
) -> t.Union[T, None]:
    ...


@t.overload
def find(
    collection: t.List[T],
    predicate: t.Union[t.Callable[[T, int], t.Any], IterateeObjT, None] = None,
) -> t.Union[T, None]:
    ...


@t.overload
def find(
    collection: t.List[T],
    predicate: t.Union[t.Callable[[T], t.Any], IterateeObjT, None] = None,
) -> t.Union[T, None]:
    ...


def find(collection, predicate=None):
    """
    Iterates over elements of a collection, returning the first element that the predicate returns
    truthy for.

    Args:
        collection: Collection to iterate over.
        predicate: Predicate applied per iteration.

    Returns:
        First element found or ``None``.

    Example:

        >>> find([1, 2, 3, 4], lambda x: x >= 3)
        3
        >>> find([{'a': 1}, {'b': 2}, {'a': 1, 'b': 2}], {'a': 1})
        {'a': 1}

    .. versionadded:: 1.0.0

    .. versionchanged:: 4.0.0
        Removed aliases ``detect`` and ``find_where``.
    """
    search = (value for is_true, value, _, _ in iteriteratee(collection, predicate) if is_true)
    return next(search, None)


@t.overload
def find_last(
    collection: t.Dict[T, T2],
    predicate: t.Union[t.Callable[[T2, T, t.Dict[T, T2]], t.Any], IterateeObjT, None] = None,
) -> t.Union[T2, None]:
    ...


@t.overload
def find_last(
    collection: t.Dict[T, T2],
    predicate: t.Union[t.Callable[[T2, T], t.Any], IterateeObjT, None] = None,
) -> t.Union[T2, None]:
    ...


@t.overload
def find_last(
    collection: t.Dict[t.Any, T2],
    predicate: t.Union[t.Callable[[T2], t.Any], IterateeObjT, None] = None,
) -> t.Union[T2, None]:
    ...


@t.overload
def find_last(
    collection: t.List[T],
    predicate: t.Union[t.Callable[[T, int, t.List[T]], t.Any], IterateeObjT, None] = None,
) -> t.Union[T, None]:
    ...


@t.overload
def find_last(
    collection: t.List[T],
    predicate: t.Union[t.Callable[[T, int], t.Any], IterateeObjT, None] = None,
) -> t.Union[T, None]:
    ...


@t.overload
def find_last(
    collection: t.List[T],
    predicate: t.Union[t.Callable[[T], t.Any], IterateeObjT, None] = None,
) -> t.Union[T, None]:
    ...


def find_last(collection, predicate=None):
    """
    This method is like :func:`find` except that it iterates over elements of a `collection` from
    right to left.

    Args:
        collection: Collection to iterate over.
        predicate: Predicate applied per iteration.

    Returns:
        Last element found or ``None``.

    Example:

        >>> find_last([1, 2, 3, 4], lambda x: x >= 3)
        4
        >>> results = find_last([{'a': 1}, {'b': 2}, {'a': 1, 'b': 2}],\
                                 {'a': 1})
        >>> assert results == {'a': 1, 'b': 2}

    .. versionadded:: 1.0.0
    """
    search = (
        value
        for is_true, value, _, _ in iteriteratee(collection, predicate, reverse=True)
        if is_true
    )
    return next(search, None)


@t.overload
def flat_map(
    collection: t.Mapping[T, T2], iteratee: t.Callable[[T2, T, t.Dict[T, T2]], t.Iterable[T3]]
) -> t.List[T3]:
    ...


@t.overload
def flat_map(
    collection: t.Mapping[T, T2], iteratee: t.Callable[[T2, T], t.Iterable[T3]]
) -> t.List[T3]:
    ...


@t.overload
def flat_map(
    collection: t.Mapping[t.Any, T2], iteratee: t.Callable[[T2], t.Iterable[T3]]
) -> t.List[T3]:
    ...


@t.overload
def flat_map(
    collection: t.Mapping[T, T2], iteratee: t.Callable[[T2, T, t.Dict[T, T2]], T3]
) -> t.List[T3]:
    ...


@t.overload
def flat_map(collection: t.Mapping[T, T2], iteratee: t.Callable[[T2, T], T3]) -> t.List[T3]:
    ...


@t.overload
def flat_map(collection: t.Mapping[t.Any, T2], iteratee: t.Callable[[T2], T3]) -> t.List[T3]:
    ...


@t.overload
def flat_map(collection: t.Mapping[t.Any, t.Iterable[T2]], iteratee: None = None) -> t.List[T2]:
    ...


@t.overload
def flat_map(collection: t.Mapping[t.Any, T2], iteratee: None = None) -> t.List[T2]:
    ...


@t.overload
def flat_map(
    collection: t.Iterable[T], iteratee: t.Callable[[T, int, t.List[T]], t.Iterable[T2]]
) -> t.List[T2]:
    ...


@t.overload
def flat_map(
    collection: t.Iterable[T], iteratee: t.Callable[[T, int], t.Iterable[T2]]
) -> t.List[T2]:
    ...


@t.overload
def flat_map(collection: t.Iterable[T], iteratee: t.Callable[[T], t.Iterable[T2]]) -> t.List[T2]:
    ...


@t.overload
def flat_map(
    collection: t.Iterable[T], iteratee: t.Callable[[T, int, t.List[T]], T2]
) -> t.List[T2]:
    ...


@t.overload
def flat_map(collection: t.Iterable[T], iteratee: t.Callable[[T, int], T2]) -> t.List[T2]:
    ...


@t.overload
def flat_map(collection: t.Iterable[T], iteratee: t.Callable[[T], T2]) -> t.List[T2]:
    ...


@t.overload
def flat_map(collection: t.Iterable[t.Iterable[T]], iteratee: None = None) -> t.List[T]:
    ...


@t.overload
def flat_map(collection: t.Iterable[T], iteratee: None = None) -> t.List[T]:
    ...


def flat_map(collection, iteratee=None):
    """
    Creates a flattened list of values by running each element in collection through `iteratee` and
    flattening the mapped results. The `iteratee` is invoked with three arguments: ``(value,
    index|key, collection)``.

    Args:
        collection: Collection to iterate over.
        iteratee: Iteratee applied per iteration.

    Returns:
        Flattened mapped list.

    Example:

        >>> duplicate = lambda n: [[n, n]]
        >>> flat_map([1, 2], duplicate)
        [[1, 1], [2, 2]]

    .. versionadded:: 4.0.0
    """
    return pyd.flatten(itermap(collection, iteratee=iteratee))


@t.overload
def flat_map_deep(
    collection: t.Mapping[T, T2],
    iteratee: t.Union[t.Callable[[T2, T, t.Dict[T, T2]], t.Any], None] = None,
) -> t.List:
    ...


@t.overload
def flat_map_deep(
    collection: t.Mapping[T, T2], iteratee: t.Union[t.Callable[[T2, T], t.Any], None] = None
) -> t.List:
    ...


@t.overload
def flat_map_deep(
    collection: t.Mapping[t.Any, T2], iteratee: t.Union[t.Callable[[T2], t.Any], None] = None
) -> t.List:
    ...


@t.overload
def flat_map_deep(
    collection: t.Iterable[T],
    iteratee: t.Union[t.Callable[[T, int, t.List[T]], t.Any], None] = None,
) -> t.List:
    ...


@t.overload
def flat_map_deep(
    collection: t.Iterable[T], iteratee: t.Union[t.Callable[[T, int], t.Any], None] = None
) -> t.List:
    ...


@t.overload
def flat_map_deep(
    collection: t.Iterable[T], iteratee: t.Union[t.Callable[[T], t.Any], None] = None
) -> t.List:
    ...


def flat_map_deep(collection: t.Iterable, iteratee=None):
    """
    This method is like :func:`flat_map` except that it recursively flattens the mapped results.

    Args:
        collection: Collection to iterate over.
        iteratee: Iteratee applied per iteration.

    Returns:
        Flattened mapped list.

    Example:

        >>> duplicate = lambda n: [[n, n]]
        >>> flat_map_deep([1, 2], duplicate)
        [1, 1, 2, 2]

    .. versionadded:: 4.0.0
    """
    return pyd.flatten_deep(itermap(collection, iteratee=iteratee))


@t.overload
def flat_map_depth(
    collection: t.Mapping[T, T2],
    iteratee: t.Union[t.Callable[[T2, T, t.Dict[T, T2]], t.Any], None] = None,
    depth: int = 1,
) -> t.List:
    ...


@t.overload
def flat_map_depth(
    collection: t.Mapping[T, T2],
    iteratee: t.Union[t.Callable[[T2, T], t.Any], None] = None,
    depth: int = 1,
) -> t.List:
    ...


@t.overload
def flat_map_depth(
    collection: t.Mapping[t.Any, T2],
    iteratee: t.Union[t.Callable[[T2], t.Any], None] = None,
    depth: int = 1,
) -> t.List:
    ...


@t.overload
def flat_map_depth(
    collection: t.Iterable[T],
    iteratee: t.Union[t.Callable[[T, int, t.List[T]], t.Any], None] = None,
    depth: int = 1,
) -> t.List:
    ...


@t.overload
def flat_map_depth(
    collection: t.Iterable[T],
    iteratee: t.Union[t.Callable[[T, int], t.Any], None] = None,
    depth: int = 1,
) -> t.List:
    ...


@t.overload
def flat_map_depth(
    collection: t.Iterable[T],
    iteratee: t.Union[t.Callable[[T], t.Any], None] = None,
    depth: int = 1,
) -> t.List:
    ...


def flat_map_depth(collection, iteratee=None, depth=1):
    """
    This method is like :func:`flat_map` except that it recursively flattens the mapped results up
    to `depth` times.

    Args:
        collection: Collection to iterate over.
        iteratee: Iteratee applied per iteration.

    Returns:
        Flattened mapped list.

    Example:

        >>> duplicate = lambda n: [[n, n]]
        >>> flat_map_depth([1, 2], duplicate, 1)
        [[1, 1], [2, 2]]
        >>> flat_map_depth([1, 2], duplicate, 2)
        [1, 1, 2, 2]

    .. versionadded:: 4.0.0
    """
    return pyd.flatten_depth(itermap(collection, iteratee=iteratee), depth=depth)


@t.overload
def for_each(
    collection: t.Dict[T, T2],
    iteratee: t.Union[t.Callable[[T2, T, t.Dict[T, T2]], t.Any], IterateeObjT, None] = None,
) -> t.Dict[T, T2]:
    ...


@t.overload
def for_each(
    collection: t.Dict[T, T2],
    iteratee: t.Union[t.Callable[[T2, T], t.Any], IterateeObjT, None] = None,
) -> t.Dict[T, T2]:
    ...


@t.overload
def for_each(
    collection: t.Dict[T, T2],
    iteratee: t.Union[t.Callable[[T2], t.Any], IterateeObjT, None] = None,
) -> t.Dict[T, T2]:
    ...


@t.overload
def for_each(
    collection: t.List[T],
    iteratee: t.Union[t.Callable[[T, int, t.List[T]], t.Any], IterateeObjT, None] = None,
) -> t.List[T]:
    ...


@t.overload
def for_each(
    collection: t.List[T],
    iteratee: t.Union[t.Callable[[T, int], t.Any], IterateeObjT, None] = None,
) -> t.List[T]:
    ...


@t.overload
def for_each(
    collection: t.List[T],
    iteratee: t.Union[t.Callable[[T], t.Any], IterateeObjT, None] = None,
) -> t.List[T]:
    ...


def for_each(collection, iteratee=None):
    """
    Iterates over elements of a collection, executing the iteratee for each element.

    Args:
        collection: Collection to iterate over.
        iteratee: Iteratee applied per iteration.

    Returns:
        `collection`

    Example:

        >>> results = {}
        >>> def cb(x): results[x] = x ** 2
        >>> for_each([1, 2, 3, 4], cb)
        [1, 2, 3, 4]
        >>> assert results == {1: 1, 2: 4, 3: 9, 4: 16}

    .. versionadded:: 1.0.0

    .. versionchanged:: 4.0.0
        Removed alias ``each``.
    """
    next((None for ret, _, _, _ in iteriteratee(collection, iteratee) if ret is False), None)
    return collection


@t.overload
def for_each_right(
    collection: t.Dict[T, T2],
    iteratee: t.Union[t.Callable[[T2, T, t.Dict[T, T2]], t.Any], IterateeObjT],
) -> t.Dict[T, T2]:
    ...


@t.overload
def for_each_right(
    collection: t.Dict[T, T2],
    iteratee: t.Union[t.Callable[[T2, T], t.Any], IterateeObjT],
) -> t.Dict[T, T2]:
    ...


@t.overload
def for_each_right(
    collection: t.Dict[T, T2],
    iteratee: t.Union[t.Callable[[T2], t.Any], IterateeObjT],
) -> t.Dict[T, T2]:
    ...


@t.overload
def for_each_right(
    collection: t.List[T],
    iteratee: t.Union[t.Callable[[T, int, t.List[T]], t.Any], IterateeObjT],
) -> t.List[T]:
    ...


@t.overload
def for_each_right(
    collection: t.List[T],
    iteratee: t.Union[t.Callable[[T, int], t.Any], IterateeObjT],
) -> t.List[T]:
    ...


@t.overload
def for_each_right(
    collection: t.List[T],
    iteratee: t.Union[t.Callable[[T], t.Any], IterateeObjT],
) -> t.List[T]:
    ...


def for_each_right(collection, iteratee):
    """
    This method is like :func:`for_each` except that it iterates over elements of a `collection`
    from right to left.

    Args:
        collection: Collection to iterate over.
        iteratee: Iteratee applied per iteration.

    Returns:
        `collection`

    Example:

        >>> results = {'total': 1}
        >>> def cb(x): results['total'] = x * results['total']
        >>> for_each_right([1, 2, 3, 4], cb)
        [1, 2, 3, 4]
        >>> assert results == {'total': 24}

    .. versionadded:: 1.0.0

    .. versionchanged:: 4.0.0
        Removed alias ``each_right``.
    """
    next(
        (None for ret, _, _, _ in iteriteratee(collection, iteratee, reverse=True) if ret is False),
        None,
    )
    return collection


@t.overload
def group_by(collection: t.Iterable[T], iteratee: t.Callable[[T], T2]) -> t.Dict[T2, t.List[T]]:
    ...


@t.overload
def group_by(
    collection: t.Iterable[T], iteratee: t.Union[IterateeObjT, None] = None
) -> t.Dict[t.Any, t.List[T]]:
    ...


def group_by(collection, iteratee=None):
    """
    Creates an object composed of keys generated from the results of running each element of a
    `collection` through the iteratee.

    Args:
        collection: Collection to iterate over.
        iteratee: Iteratee applied per iteration.

    Returns:
        Results of grouping by `iteratee`.

    Example:

        >>> results = group_by([{'a': 1, 'b': 2}, {'a': 3, 'b': 4}], 'a')
        >>> assert results == {1: [{'a': 1, 'b': 2}], 3: [{'a': 3, 'b': 4}]}
        >>> results = group_by([{'a': 1, 'b': 2}, {'a': 3, 'b': 4}], {'a': 1})
        >>> assert results == {False: [{'a': 3, 'b': 4}],\
                               True: [{'a': 1, 'b': 2}]}

    .. versionadded:: 1.0.0
    """
    ret = {}
    cbk = pyd.iteratee(iteratee)

    for value in collection:
        key = cbk(value)
        ret.setdefault(key, [])
        ret[key].append(value)

    return ret


def includes(collection: t.Union[t.Sequence, t.Dict], target: t.Any, from_index: int = 0) -> bool:
    """
    Checks if a given value is present in a collection. If `from_index` is negative, it is used as
    the offset from the end of the collection.

    Args:
        collection: Collection to iterate over.
        target: Target value to compare to.
        from_index: Offset to start search from.

    Returns:
        Whether `target` is in `collection`.

    Example:

        >>> includes([1, 2, 3, 4], 2)
        True
        >>> includes([1, 2, 3, 4], 2, from_index=2)
        False
        >>> includes({'a': 1, 'b': 2, 'c': 3, 'd': 4}, 2)
        True

    .. versionadded:: 1.0.0

    .. versionchanged:: 4.0.0
        Renamed from ``contains`` to ``includes`` and removed alias
        ``include``.
    """
    collection_values: t.Container
    if isinstance(collection, dict):
        collection_values = collection.values()
    else:
        # only makes sense to do this if `collection` is not a dict
        collection_values = collection[from_index:]

    return target in collection_values


def invoke_map(collection: t.Iterable, path: PathT, *args: t.Any, **kwargs: t.Any) -> t.List[t.Any]:
    """
    Invokes the method at `path` of each element in `collection`, returning a list of the results of
    each invoked method. Any additional arguments are provided to each invoked method. If `path` is
    a function, it's invoked for each element in `collection`.

    Args:
        collection: Collection to iterate over.
        path: String path to method to invoke or callable to invoke for each element in
            `collection`.
        args: Arguments to pass to method call.
        kwargs: Keyword arguments to pass to method call.

    Returns:
        List of results of invoking method of each item.

    Example:

        >>> items = [{'a': [{'b': 1}]}, {'a': [{'c': 2}]}]
        >>> expected = [{'b': 1}.items(), {'c': 2}.items()]
        >>> invoke_map(items, 'a[0].items') == expected
        True

    .. versionadded:: 4.0.0
    """
    return map_(collection, lambda item: pyd.invoke(item, path, *args, **kwargs))


@t.overload
def key_by(collection: t.Iterable[T], iteratee: t.Callable[[T], T2]) -> t.Dict[T2, T]:
    ...


@t.overload
def key_by(collection: t.Iterable, iteratee: t.Union[IterateeObjT, None] = None) -> t.Dict:
    ...


def key_by(collection, iteratee=None):
    """
    Creates an object composed of keys generated from the results of running each element of the
    collection through the given iteratee.

    Args:
        collection: Collection to iterate over.
        iteratee: Iteratee applied per iteration.

    Returns:
        Results of indexing by `iteratee`.

    Example:

        >>> results = key_by([{'a': 1, 'b': 2}, {'a': 3, 'b': 4}], 'a')
        >>> assert results == {1: {'a': 1, 'b': 2}, 3: {'a': 3, 'b': 4}}


    .. versionadded:: 1.0.0

    .. versionchanged:: 4.0.0
        Renamed from ``index_by`` to ``key_by``.
    """
    ret = {}
    cbk = pyd.iteratee(iteratee)

    for value in collection:
        ret[cbk(value)] = value

    return ret


@t.overload
def map_(collection: t.Mapping[t.Any, T2], iteratee: t.Callable[[T2], T3]) -> t.List[T3]:
    ...


@t.overload
def map_(collection: t.Mapping[T, T2], iteratee: t.Callable[[T2, T], T3]) -> t.List[T3]:
    ...


@t.overload
def map_(
    collection: t.Mapping[T, T2], iteratee: t.Callable[[T2, T, t.Dict[T, T2]], T3]
) -> t.List[T3]:
    ...


@t.overload
def map_(collection: t.Iterable[T], iteratee: t.Callable[[T], T2]) -> t.List[T2]:
    ...


@t.overload
def map_(collection: t.Iterable[T], iteratee: t.Callable[[T, int], T2]) -> t.List[T2]:
    ...


@t.overload
def map_(collection: t.Iterable[T], iteratee: t.Callable[[T, int, t.List[T]], T2]) -> t.List[T2]:
    ...


@t.overload
def map_(collection: t.Iterable, iteratee: t.Union[IterateeObjT, None] = None) -> t.List:
    ...


def map_(collection, iteratee=None):
    """
    Creates an array of values by running each element in the collection through the iteratee. The
    iteratee is invoked with three arguments: ``(value, index|key, collection)``. If a property name
    is passed for iteratee, the created :func:`pluck` style iteratee will return the property value
    of the given element. If an object is passed for iteratee, the created :func:`.matches` style
    iteratee will return ``True`` for elements that have the properties of the given object, else
    ``False``.

    Args:
        collection: Collection to iterate over.
        iteratee: Iteratee applied per iteration.

    Returns:
        Mapped list.

    Example:

        >>> map_([1, 2, 3, 4], str)
        ['1', '2', '3', '4']
        >>> map_([{'a': 1, 'b': 2}, {'a': 3, 'b': 4}, {'a': 5, 'b': 6}], 'a')
        [1, 3, 5]
        >>> map_([[[0, 1]], [[2, 3]], [[4, 5]]], '0.1')
        [1, 3, 5]
        >>> map_([{'a': {'b': 1}}, {'a': {'b': 2}}], 'a.b')
        [1, 2]
        >>> map_([{'a': {'b': [0, 1]}}, {'a': {'b': [2, 3]}}], 'a.b[1]')
        [1, 3]

    .. versionadded:: 1.0.0

    .. versionchanged:: 4.0.0
        Removed alias ``collect``.
    """
    return list(itermap(collection, iteratee))


def nest(collection: t.Iterable, *properties: t.Any) -> t.Any:
    """
    This method is like :func:`group_by` except that it supports nested grouping by multiple string
    `properties`. If only a single key is given, it is like calling ``group_by(collection, prop)``.

    Args:
        collection: Collection to iterate over.
        *properties: Properties to nest by.

    Returns:
        Results of nested grouping by `properties`.

    Example:

        >>> results = nest([{'shape': 'square', 'color': 'red', 'qty': 5},\
                            {'shape': 'square', 'color': 'blue', 'qty': 10},\
                            {'shape': 'square', 'color': 'orange', 'qty': 5},\
                            {'shape': 'circle', 'color': 'yellow', 'qty': 5},\
                            {'shape': 'circle', 'color': 'pink', 'qty': 10},\
                            {'shape': 'oval', 'color': 'purple', 'qty': 5}],\
                           'shape', 'qty')
        >>> expected = {\
            'square': {5: [{'shape': 'square', 'color': 'red', 'qty': 5},\
                           {'shape': 'square', 'color': 'orange', 'qty': 5}],\
                       10: [{'shape': 'square', 'color': 'blue', 'qty': 10}]},\
            'circle': {5: [{'shape': 'circle', 'color': 'yellow', 'qty': 5}],\
                       10: [{'shape': 'circle', 'color': 'pink', 'qty': 10}]},\
            'oval': {5: [{'shape': 'oval', 'color': 'purple', 'qty': 5}]}}
        >>> results == expected
        True

    .. versionadded:: 4.3.0
    """
    if not properties:
        return collection

    flat_properties = pyd.flatten(properties)
    first, rest = flat_properties[0], flat_properties[1:]

    return pyd.map_values(group_by(collection, first), lambda value: nest(value, *rest))


@t.overload
def order_by(
    collection: t.Mapping[t.Any, T2],
    keys: t.Iterable[t.Union[str, int]],
    orders: t.Union[t.Iterable[bool], bool],
    reverse: bool = False,
) -> t.List[T2]:
    ...


@t.overload
def order_by(
    collection: t.Mapping[t.Any, T2],
    keys: t.Iterable[str],
    orders: None = None,
    reverse: bool = False,
) -> t.List[T2]:
    ...


@t.overload
def order_by(
    collection: t.Iterable[T],
    keys: t.Iterable[t.Union[str, int]],
    orders: t.Union[t.Iterable[bool], bool],
    reverse: bool = False,
) -> t.List[T]:
    ...


@t.overload
def order_by(
    collection: t.Iterable[T],
    keys: t.Iterable[str],
    orders: None = None,
    reverse: bool = False,
) -> t.List[T]:
    ...


def order_by(collection, keys, orders=None, reverse=False):
    """
    This method is like :func:`sort_by` except that it sorts by key names instead of an iteratee
    function. Keys can be sorted in descending order by prepending a ``"-"`` to the key name (e.g.
    ``"name"`` would become ``"-name"``) or by passing a list of boolean sort options via `orders`
    where ``True`` is ascending and ``False`` is descending.

    Args:
        collection: Collection to iterate over.
        keys: List of keys to sort by. By default, keys will be sorted in ascending order. To
            sort a key in descending order, prepend a ``"-"`` to the key name. For example, to sort
            the key value for ``"name"`` in descending order, use ``"-name"``.
        orders: List of boolean sort orders to apply for each key. ``True``
            corresponds to ascending order while ``False`` is descending. Defaults to ``None``.
        reverse (bool, optional): Whether to reverse the sort. Defaults to ``False``.

    Returns:
        Sorted list.

    Example:

        >>> items = [{'a': 2, 'b': 1}, {'a': 3, 'b': 2}, {'a': 1, 'b': 3}]
        >>> results = order_by(items, ['b', 'a'])
        >>> assert results == [{'a': 2, 'b': 1},\
                               {'a': 3, 'b': 2},\
                               {'a': 1, 'b': 3}]
        >>> results = order_by(items, ['a', 'b'])
        >>> assert results == [{'a': 1, 'b': 3},\
                               {'a': 2, 'b': 1},\
                               {'a': 3, 'b': 2}]
        >>> results = order_by(items, ['-a', 'b'])
        >>> assert results == [{'a': 3, 'b': 2},\
                               {'a': 2, 'b': 1},\
                               {'a': 1, 'b': 3}]
        >>> results = order_by(items, ['a', 'b'], [False, True])
        >>> assert results == [{'a': 3, 'b': 2},\
                               {'a': 2, 'b': 1},\
                               {'a': 1, 'b': 3}]

    .. versionadded:: 3.0.0

    .. versionchanged:: 3.2.0
        Added `orders` argument.

    .. versionchanged:: 3.2.0
        Added :func:`sort_by_order` as alias.

    .. versionchanged:: 4.0.0
        Renamed from ``order_by`` to ``order_by`` and removed alias
        ``sort_by_order``.
    """
    if isinstance(collection, dict):
        collection = collection.values()

    # Maintain backwards compatibility.
    if pyd.is_boolean(orders):
        reverse = orders
        orders = None

    comparers = []

    if orders:
        for i, key in enumerate(keys):
            if pyd.has(orders, i):
                order = 1 if orders[i] else -1
            else:
                order = 1

            comparers.append((pyd.property_(key), order))
    else:
        for key in keys:
            if key.startswith("-"):
                order = -1
                key = key[1:]
            else:
                order = 1

            comparers.append((pyd.property_(key), order))

    def comparison(left, right):
        # pylint: disable=useless-else-on-loop,missing-docstring
        for func, mult in comparers:
            result = cmp(func(left), func(right))
            if result:
                return mult * result
        else:
            return 0

    return sorted(collection, key=cmp_to_key(comparison), reverse=reverse)


@t.overload
def partition(
    collection: t.Mapping[T, T2], predicate: t.Callable[[T2, T, t.Dict[T, T2]], t.Any]
) -> t.List[t.List[T2]]:
    ...


@t.overload
def partition(
    collection: t.Mapping[T, T2], predicate: t.Callable[[T2, T], t.Any]
) -> t.List[t.List[T2]]:
    ...


@t.overload
def partition(
    collection: t.Mapping[t.Any, T2], predicate: t.Callable[[T2], t.Any]
) -> t.List[t.List[T2]]:
    ...


@t.overload
def partition(
    collection: t.Mapping[t.Any, T2], predicate: t.Union[IterateeObjT, None] = None
) -> t.List[t.List[T2]]:
    ...


@t.overload
def partition(
    collection: t.Iterable[T], predicate: t.Callable[[T, int, t.List[T]], t.Any]
) -> t.List[t.List[T]]:
    ...


@t.overload
def partition(
    collection: t.Iterable[T], predicate: t.Callable[[T, int], t.Any]
) -> t.List[t.List[T]]:
    ...


@t.overload
def partition(collection: t.Iterable[T], predicate: t.Callable[[T], t.Any]) -> t.List[t.List[T]]:
    ...


@t.overload
def partition(
    collection: t.Iterable[T], predicate: t.Union[IterateeObjT, None] = None
) -> t.List[t.List[T]]:
    ...


def partition(collection, predicate=None):
    """
    Creates an array of elements split into two groups, the first of which contains elements the
    `predicate` returns truthy for, while the second of which contains elements the `predicate`
    returns falsey for. The `predicate` is invoked with three arguments: ``(value, index|key,
    collection)``.

    If a property name is provided for `predicate` the created :func:`pluck` style predicate returns
    the property value of the given element.

    If an object is provided for `predicate` the created :func:`.matches` style predicate returns
    ``True`` for elements that have the properties of the given object, else ``False``.

    Args:
        collection: Collection to iterate over.
        predicate: Predicate applied per iteration.

    Returns:
        List of grouped elements.

    Example:

        >>> partition([1, 2, 3, 4], lambda x: x >= 3)
        [[3, 4], [1, 2]]

    .. versionadded:: 1.1.0
    """
    trues = []
    falses = []

    for is_true, value, _, _ in iteriteratee(collection, predicate):
        if is_true:
            trues.append(value)
        else:
            falses.append(value)

    return [trues, falses]


def pluck(collection: t.Iterable, path: PathT) -> t.List:
    """
    Retrieves the value of a specified property from all elements in the collection.

    Args:
        collection: List of dicts.
        path: Collection's path to pluck

    Returns:
        Plucked list.

    Example:

        >>> pluck([{'a': 1, 'b': 2}, {'a': 3, 'b': 4}, {'a': 5, 'b': 6}], 'a')
        [1, 3, 5]
        >>> pluck([[[0, 1]], [[2, 3]], [[4, 5]]], '0.1')
        [1, 3, 5]
        >>> pluck([{'a': {'b': 1}}, {'a': {'b': 2}}], 'a.b')
        [1, 2]
        >>> pluck([{'a': {'b': [0, 1]}}, {'a': {'b': [2, 3]}}], 'a.b.1')
        [1, 3]
        >>> pluck([{'a': {'b': [0, 1]}}, {'a': {'b': [2, 3]}}], ['a', 'b', 1])
        [1, 3]

    .. versionadded:: 1.0.0

    .. versionchanged:: 4.0.0
        Function removed.

    .. versionchanged:: 4.0.1
        Made property access deep.
    """
    return map_(collection, pyd.property_(path))


@t.overload
def reduce_(
    collection: t.Mapping[T, T2],
    iteratee: t.Callable[[T3, T2, T], T3],
    accumulator: T3,
) -> T3:
    ...


@t.overload
def reduce_(
    collection: t.Mapping[t.Any, T2],
    iteratee: t.Callable[[T3, T2], T3],
    accumulator: T3,
) -> T3:
    ...


@t.overload
def reduce_(
    collection: t.Mapping,
    iteratee: t.Callable[[T3], T3],
    accumulator: T3,
) -> T3:
    ...


@t.overload
def reduce_(
    collection: t.Mapping[T, T2],
    iteratee: t.Callable[[T2, T2, T], T2],
    accumulator: None = None,
) -> T2:
    ...


@t.overload
def reduce_(
    collection: t.Mapping[t.Any, T2],
    iteratee: t.Callable[[T2, T2], T2],
    accumulator: None = None,
) -> T2:
    ...


@t.overload
def reduce_(
    collection: t.Mapping,
    iteratee: t.Callable[[T], T],
    accumulator: None = None,
) -> T:
    ...


@t.overload
def reduce_(
    collection: t.Iterable[T],
    iteratee: t.Callable[[T2, T, int], T2],
    accumulator: T2,
) -> T2:
    ...


@t.overload
def reduce_(
    collection: t.Iterable[T],
    iteratee: t.Callable[[T2, T], T2],
    accumulator: T2,
) -> T2:
    ...


@t.overload
def reduce_(
    collection: t.Iterable,
    iteratee: t.Callable[[T2], T2],
    accumulator: T2,
) -> T2:
    ...


@t.overload
def reduce_(
    collection: t.Iterable[T],
    iteratee: t.Callable[[T, T, int], T],
    accumulator: None = None,
) -> T:
    ...


@t.overload
def reduce_(
    collection: t.Iterable[T],
    iteratee: t.Callable[[T, T], T],
    accumulator: None = None,
) -> T:
    ...


@t.overload
def reduce_(
    collection: t.Iterable,
    iteratee: t.Callable[[T], T],
    accumulator: None = None,
) -> T:
    ...


@t.overload
def reduce_(
    collection: t.Iterable[T], iteratee: None = None, accumulator: t.Union[T, None] = None
) -> T:
    ...


def reduce_(collection, iteratee=None, accumulator=None):
    """
    Reduces a collection to a value which is the accumulated result of running each element in the
    collection through the iteratee, where each successive iteratee execution consumes the return
    value of the previous execution.

    Args:
        collection: Collection to iterate over.
        iteratee: Iteratee applied per iteration.
        accumulator: Initial value of aggregator. Default is to use the result of
            the first iteration.

    Returns:
        Accumulator object containing results of reduction.

    Example:

        >>> reduce_([1, 2, 3, 4], lambda total, x: total * x)
        24

    .. versionadded:: 1.0.0

    .. versionchanged:: 4.0.0
        Removed aliases ``foldl`` and ``inject``.
    """
    iterable = iterator(collection)

    if accumulator is None:
        try:
            _, accumulator = next(iterable)
        except StopIteration:
            raise TypeError("reduce_() of empty sequence with no initial value")

    result = accumulator

    if iteratee is None:
        iteratee = pyd.identity
        argcount = 1
    else:
        argcount = getargcount(iteratee, maxargs=3)

    for index, item in iterable:
        result = callit(iteratee, result, item, index, argcount=argcount)

    return result


@t.overload
def reduce_right(
    collection: t.Mapping[T, T2],
    iteratee: t.Callable[[T3, T2, T], T3],
    accumulator: T3,
) -> T3:
    ...


@t.overload
def reduce_right(
    collection: t.Mapping[t.Any, T2],
    iteratee: t.Callable[[T3, T2], T3],
    accumulator: T3,
) -> T3:
    ...


@t.overload
def reduce_right(
    collection: t.Mapping,
    iteratee: t.Callable[[T3], T3],
    accumulator: T3,
) -> T3:
    ...


@t.overload
def reduce_right(
    collection: t.Mapping[T, T2],
    iteratee: t.Callable[[T2, T2, T], T2],
    accumulator: None = None,
) -> T2:
    ...


@t.overload
def reduce_right(
    collection: t.Mapping[t.Any, T2],
    iteratee: t.Callable[[T2, T2], T2],
    accumulator: None = None,
) -> T2:
    ...


@t.overload
def reduce_right(
    collection: t.Mapping,
    iteratee: t.Callable[[T], T],
    accumulator: None = None,
) -> T:
    ...


@t.overload
def reduce_right(
    collection: t.Iterable[T],
    iteratee: t.Callable[[T2, T, int], T2],
    accumulator: T2,
) -> T2:
    ...


@t.overload
def reduce_right(
    collection: t.Iterable[T],
    iteratee: t.Callable[[T2, T], T2],
    accumulator: T2,
) -> T2:
    ...


@t.overload
def reduce_right(
    collection: t.Iterable,
    iteratee: t.Callable[[T2], T2],
    accumulator: T2,
) -> T2:
    ...


@t.overload
def reduce_right(
    collection: t.Iterable[T],
    iteratee: t.Callable[[T, T, int], T],
    accumulator: None = None,
) -> T:
    ...


@t.overload
def reduce_right(
    collection: t.Iterable[T],
    iteratee: t.Callable[[T, T], T],
    accumulator: None = None,
) -> T:
    ...


@t.overload
def reduce_right(
    collection: t.Iterable,
    iteratee: t.Callable[[T], T],
    accumulator: None = None,
) -> T:
    ...


@t.overload
def reduce_right(
    collection: t.Iterable[T], iteratee: None = None, accumulator: t.Union[T, None] = None
) -> T:
    ...


def reduce_right(collection, iteratee=None, accumulator=None):
    """
    This method is like :func:`reduce_` except that it iterates over elements of a `collection` from
    right to left.

    Args:
        collection: Collection to iterate over.
        iteratee: Iteratee applied per iteration.
        accumulator: Initial value of aggregator. Default is to use the result of
            the first iteration.

    Returns:
        Accumulator object containing results of reduction.

    Example:

        >>> reduce_right([1, 2, 3, 4], lambda total, x: total ** x)
        4096

    .. versionadded:: 1.0.0

    .. versionchanged:: 3.2.1
        Fix bug where collection was not reversed correctly.

    .. versionchanged:: 4.0.0
        Removed alias ``foldr``.
    """
    if not isinstance(collection, dict):
        collection = list(collection)[::-1]

    return reduce_(collection, iteratee, accumulator)


@t.overload
def reductions(
    collection: t.Mapping[T, T2],
    iteratee: t.Callable[[T3, T2, T], T3],
    accumulator: T3,
    from_right: bool = False,
) -> t.List[T3]:
    ...


@t.overload
def reductions(
    collection: t.Mapping[t.Any, T2],
    iteratee: t.Callable[[T3, T2], T3],
    accumulator: T3,
    from_right: bool = False,
) -> t.List[T3]:
    ...


@t.overload
def reductions(
    collection: t.Mapping,
    iteratee: t.Callable[[T3], T3],
    accumulator: T3,
    from_right: bool = False,
) -> t.List[T3]:
    ...


@t.overload
def reductions(
    collection: t.Mapping[T, T2],
    iteratee: t.Callable[[T2, T2, T], T2],
    accumulator: None = None,
    from_right: bool = False,
) -> t.List[T2]:
    ...


@t.overload
def reductions(
    collection: t.Mapping[t.Any, T2],
    iteratee: t.Callable[[T2, T2], T2],
    accumulator: None = None,
    from_right: bool = False,
) -> t.List[T2]:
    ...


@t.overload
def reductions(
    collection: t.Mapping,
    iteratee: t.Callable[[T], T],
    accumulator: None = None,
    from_right: bool = False,
) -> t.List[T]:
    ...


@t.overload
def reductions(
    collection: t.Iterable[T],
    iteratee: t.Callable[[T2, T, int], T2],
    accumulator: T2,
    from_right: bool = False,
) -> t.List[T2]:
    ...


@t.overload
def reductions(
    collection: t.Iterable[T],
    iteratee: t.Callable[[T2, T], T2],
    accumulator: T2,
    from_right: bool = False,
) -> t.List[T2]:
    ...


@t.overload
def reductions(
    collection: t.Iterable,
    iteratee: t.Callable[[T2], T2],
    accumulator: T2,
    from_right: bool = False,
) -> t.List[T2]:
    ...


@t.overload
def reductions(
    collection: t.Iterable[T],
    iteratee: t.Callable[[T, T, int], T],
    accumulator: None = None,
    from_right: bool = False,
) -> t.List[T]:
    ...


@t.overload
def reductions(
    collection: t.Iterable[T],
    iteratee: t.Callable[[T, T], T],
    accumulator: None = None,
    from_right: bool = False,
) -> t.List[T]:
    ...


@t.overload
def reductions(
    collection: t.Iterable,
    iteratee: t.Callable[[T], T],
    accumulator: None = None,
    from_right: bool = False,
) -> t.List[T]:
    ...


@t.overload
def reductions(
    collection: t.Iterable[T],
    iteratee: None = None,
    accumulator: t.Union[T, None] = None,
    from_right: bool = False,
) -> t.List[T]:
    ...


def reductions(collection, iteratee=None, accumulator=None, from_right=False):
    """
    This function is like :func:`reduce_` except that it returns a list of every intermediate value
    in the reduction operation.

    Args:
        collection: Collection to iterate over.
        iteratee: Iteratee applied per iteration.
        accumulator: Initial value of aggregator. Default is to use the result of
            the first iteration.

    Returns:
        Results of each reduction operation.

    Example:

        >>> reductions([1, 2, 3, 4], lambda total, x: total * x)
        [2, 6, 24]

    Note:
        The last element of the returned list would be the result of using
        :func:`reduce_`.

    .. versionadded:: 2.0.0
    """
    if iteratee is None:
        iteratee = pyd.identity
        argcount = 1
    else:
        argcount = getargcount(iteratee, maxargs=3)

    results = []

    def interceptor(result, item, index):
        result = callit(iteratee, result, item, index, argcount=argcount)
        results.append(result)
        return result

    reducer = reduce_right if from_right else reduce_
    reducer(collection, interceptor, accumulator)

    return results


@t.overload
def reductions_right(
    collection: t.Mapping[T, T2],
    iteratee: t.Callable[[T3, T2, T], T3],
    accumulator: T3,
) -> t.List[T3]:
    ...


@t.overload
def reductions_right(
    collection: t.Mapping[t.Any, T2],
    iteratee: t.Callable[[T3, T2], T3],
    accumulator: T3,
) -> t.List[T3]:
    ...


@t.overload
def reductions_right(
    collection: t.Mapping,
    iteratee: t.Callable[[T3], T3],
    accumulator: T3,
) -> t.List[T3]:
    ...


@t.overload
def reductions_right(
    collection: t.Mapping[T, T2],
    iteratee: t.Callable[[T2, T2, T], T2],
    accumulator: None = None,
) -> t.List[T2]:
    ...


@t.overload
def reductions_right(
    collection: t.Mapping[t.Any, T2],
    iteratee: t.Callable[[T2, T2], T2],
    accumulator: None = None,
) -> t.List[T2]:
    ...


@t.overload
def reductions_right(
    collection: t.Mapping,
    iteratee: t.Callable[[T], T],
    accumulator: None = None,
) -> t.List[T]:
    ...


@t.overload
def reductions_right(
    collection: t.Iterable[T],
    iteratee: t.Callable[[T2, T, int], T2],
    accumulator: T2,
) -> t.List[T2]:
    ...


@t.overload
def reductions_right(
    collection: t.Iterable[T],
    iteratee: t.Callable[[T2, T], T2],
    accumulator: T2,
) -> t.List[T2]:
    ...


@t.overload
def reductions_right(
    collection: t.Iterable,
    iteratee: t.Callable[[T2], T2],
    accumulator: T2,
) -> t.List[T2]:
    ...


@t.overload
def reductions_right(
    collection: t.Iterable[T],
    iteratee: t.Callable[[T, T, int], T],
    accumulator: None = None,
) -> t.List[T]:
    ...


@t.overload
def reductions_right(
    collection: t.Iterable[T],
    iteratee: t.Callable[[T, T], T],
    accumulator: None = None,
) -> t.List[T]:
    ...


@t.overload
def reductions_right(
    collection: t.Iterable,
    iteratee: t.Callable[[T], T],
    accumulator: None = None,
) -> t.List[T]:
    ...


@t.overload
def reductions_right(
    collection: t.Iterable[T], iteratee: None = None, accumulator: t.Union[T, None] = None
) -> t.List[T]:
    ...


def reductions_right(collection, iteratee=None, accumulator=None):
    """
    This method is like :func:`reductions` except that it iterates over elements of a `collection`
    from right to left.

    Args:
        collection: Collection to iterate over.
        iteratee: Iteratee applied per iteration.
        accumulator: Initial value of aggregator. Default is to use the result of
            the first iteration.

    Returns:
        Results of each reduction operation.

    Example:

        >>> reductions_right([1, 2, 3, 4], lambda total, x: total ** x)
        [64, 4096, 4096]

    Note:
        The last element of the returned list would be the result of using
        :func:`reduce_`.

    .. versionadded:: 2.0.0
    """
    return reductions(collection, iteratee, accumulator, from_right=True)


@t.overload
def reject(
    collection: t.Mapping[T, T2],
    predicate: t.Union[t.Callable[[T2, T, t.Dict[T, T2]], t.Any], IterateeObjT, None] = None,
) -> t.List[T2]:
    ...


@t.overload
def reject(
    collection: t.Mapping[T, T2],
    predicate: t.Union[t.Callable[[T2, T], t.Any], IterateeObjT, None] = None,
) -> t.List[T2]:
    ...


@t.overload
def reject(
    collection: t.Mapping[t.Any, T2],
    predicate: t.Union[t.Callable[[T2], t.Any], IterateeObjT, None] = None,
) -> t.List[T2]:
    ...


@t.overload
def reject(
    collection: t.Iterable[T],
    predicate: t.Union[t.Callable[[T, int, t.List[T]], t.Any], IterateeObjT, None] = None,
) -> t.List[T]:
    ...


@t.overload
def reject(
    collection: t.Iterable[T],
    predicate: t.Union[t.Callable[[T, int], t.Any], IterateeObjT, None] = None,
) -> t.List[T]:
    ...


@t.overload
def reject(
    collection: t.Iterable[T],
    predicate: t.Union[t.Callable[[T], t.Any], IterateeObjT, None] = None,
) -> t.List[T]:
    ...


def reject(collection, predicate=None):
    """
    The opposite of :func:`filter_` this method returns the elements of a collection that the
    predicate does **not** return truthy for.

    Args:
        collection: Collection to iterate over.
        predicate: Predicate applied per iteration.

    Returns:
        Rejected elements of `collection`.

    Example:

        >>> reject([1, 2, 3, 4], lambda x: x >= 3)
        [1, 2]
        >>> reject([{'a': 0}, {'a': 1}, {'a': 2}], 'a')
        [{'a': 0}]
        >>> reject([{'a': 0}, {'a': 1}, {'a': 2}], {'a': 1})
        [{'a': 0}, {'a': 2}]

    .. versionadded:: 1.0.0
    """
    return [value for is_true, value, _, _ in iteriteratee(collection, predicate) if not is_true]


def sample(collection: t.Sequence[T]) -> T:
    """
    Retrieves a random element from a given `collection`.

    Args:
        collection: Collection to iterate over.

    Returns:
        Random element from the given collection.

    Example:

        >>> items = [1, 2, 3, 4, 5]
        >>> results = sample(items)
        >>> assert results in items

    .. versionadded:: 1.0.0

    .. versionchanged:: 4.0.0
        Moved multiple samples functionality to :func:`sample_size`. This
        function now only returns a single random sample.
    """
    return random.choice(collection)


def sample_size(collection: t.Sequence[T], n: t.Union[int, None] = None) -> t.List[T]:
    """
    Retrieves list of `n` random elements from a collection.

    Args:
        collection: Collection to iterate over.
        n: Number of random samples to return.

    Returns:
        List of `n` sampled collection values.

    Examples:

        >>> items = [1, 2, 3, 4, 5]
        >>> results = sample_size(items, 2)
        >>> assert len(results) == 2
        >>> assert set(items).intersection(results) == set(results)

    .. versionadded:: 4.0.0
    """
    num = min(n or 1, len(collection))
    return random.sample(collection, num)


@t.overload
def shuffle(collection: t.Mapping[t.Any, T]) -> t.List[T]:
    ...


@t.overload
def shuffle(collection: t.Iterable[T]) -> t.List[T]:
    ...


def shuffle(collection):
    """
    Creates a list of shuffled values, using a version of the Fisher-Yates shuffle.

    Args:
        collection: Collection to iterate over.

    Returns:
        Shuffled list of values.

    Example:

        >>> items = [1, 2, 3, 4]
        >>> results = shuffle(items)
        >>> assert len(results) == len(items)
        >>> assert set(results) == set(items)

    .. versionadded:: 1.0.0
    """
    if isinstance(collection, dict):
        collection = collection.values()

    # Make copy of collection since random.shuffle works on list in-place.
    collection = list(collection)

    # NOTE: random.shuffle uses Fisher-Yates.
    random.shuffle(collection)

    return collection


def size(collection: t.Sized) -> int:
    """
    Gets the size of the `collection` by returning `len(collection)` for iterable objects.

    Args:
        collection: Collection to iterate over.

    Returns:
        Collection length.

    Example:

        >>> size([1, 2, 3, 4])
        4

    .. versionadded:: 1.0.0
    """
    return len(collection)


def some(
    collection: t.Iterable[T], predicate: t.Union[t.Callable[[T], t.Any], None] = None
) -> bool:
    """
    Checks if the predicate returns a truthy value for any element of a collection. The predicate is
    invoked with three arguments: ``(value, index|key, collection)``. If a property name is passed
    for predicate, the created :func:`map_` style predicate will return the property value of the
    given element. If an object is passed for predicate, the created :func:`.matches` style
    predicate will return ``True`` for elements that have the properties of the given object, else
    ``False``.

    Args:
        collection: Collection to iterate over.
        predicate: Predicate applied per iteration.

    Returns:
        Whether any of the elements are truthy.

    Example:

        >>> some([False, True, 0])
        True
        >>> some([False, 0, None])
        False
        >>> some([1, 2, 3, 4], lambda x: x >= 3)
        True
        >>> some([1, 2, 3, 4], lambda x: x == 0)
        False

    .. versionadded:: 1.0.0

    .. versionchanged:: 4.0.0
        Removed alias ``any_``.
    """
    if predicate:
        cbk = pyd.iteratee(predicate)
        collection = (cbk(item) for item in collection)

    return any(collection)


@t.overload
def sort_by(
    collection: t.Mapping[t.Any, T2],
    iteratee: t.Union[t.Callable[[T2], t.Any], IterateeObjT, None] = None,
    reverse: bool = False,
) -> t.List[T2]:
    ...


@t.overload
def sort_by(
    collection: t.Iterable[T],
    iteratee: t.Union[t.Callable[[T], t.Any], IterateeObjT, None] = None,
    reverse: bool = False,
) -> t.List[T]:
    ...


def sort_by(collection, iteratee=None, reverse=False):
    """
    Creates a list of elements, sorted in ascending order by the results of running each element in
    a `collection` through the iteratee.

    Args:
        collection: Collection to iterate over.
        iteratee: Iteratee applied per iteration.
        reverse: Whether to reverse the sort. Defaults to ``False``.

    Returns:
        Sorted list.

    Example:

        >>> sort_by({'a': 2, 'b': 3, 'c': 1})
        [1, 2, 3]
        >>> sort_by({'a': 2, 'b': 3, 'c': 1}, reverse=True)
        [3, 2, 1]
        >>> sort_by([{'a': 2}, {'a': 3}, {'a': 1}], 'a')
        [{'a': 1}, {'a': 2}, {'a': 3}]

    .. versionadded:: 1.0.0
    """
    if isinstance(collection, dict):
        collection = collection.values()

    return sorted(collection, key=pyd.iteratee(iteratee), reverse=reverse)


#
# Utility methods not a part of the main API
#


def itermap(
    collection: t.Iterable,
    iteratee: t.Union[t.Callable, IterateeObjT, None] = None,
) -> t.Generator:
    """Generative mapper."""
    for result in iteriteratee(collection, iteratee):
        yield result[0]



import datetime
import decimal
import operator
import re

import pytest

import pydash as _

from . import helpers


parametrize = pytest.mark.parametrize


@parametrize(
    "value,other,expected",
    [("a", "a", True), (None, None, True), (None, "", False), (1, str(1), False)],
)
def test_eq(value, other, expected):
    assert _.eq(value, other) == expected


@parametrize(
    "case,expected",
    [
        ((2, 1), True),
        ((2.5, 2.3), True),
        ((1, 2), False),
        ((2.3, 2.5), False),
        ((1, 1), False),
    ],
)
def test_gt(case, expected):
    assert _.gt(*case) == expected


@parametrize(
    "case,expected",
    [
        ((2, 1), True),
        ((2.5, 2.3), True),
        ((1, 2), False),
        ((2.3, 2.5), False),
        ((1, 1), True),
    ],
)
def test_gte(case, expected):
    assert _.gte(*case) == expected


@parametrize(
    "case,expected",
    [
        ((2, 1), False),
        ((2.5, 2.3), False),
        ((1, 2), True),
        ((2.3, 2.5), True),
        ((1, 1), False),
    ],
)
def test_lt(case, expected):
    assert _.lt(*case) == expected


@parametrize(
    "case,expected",
    [
        ((2, 1), False),
        ((2.5, 2.3), False),
        ((1, 2), True),
        ((2.3, 2.5), True),
        ((1, 1), True),
    ],
)
def test_lte(case, expected):
    assert _.lte(*case) == expected


@parametrize(
    "case,expected",
    [
        ((3, 2, 4), True),
        ((4, 8), True),
        ((4, 2), False),
        ((2, 2), False),
        ((1.2, 2), True),
        ((5.2, 4), False),
        (("", 5), False),
        ((2, ""), False),
        ((-1, -2, ""), True),
    ],
)
def test_in_range(case, expected):
    assert _.in_range(*case) == expected


@parametrize(
    "case,expected",
    [
        ([], True),
        ({}, True),
        ("", True),
        (0, False),
        (True, False),
    ],
)
def test_is_associative(case, expected):
    assert _.is_associative(case) == expected


@parametrize(
    "case,expected",
    [
        ("", True),
        ("\n", True),
        (" ", True),
        ("a", False),
    ],
)
def test_is_blank(case, expected):
    assert _.is_blank(case) == expected


@parametrize(
    "case,expected",
    [
        (True, True),
        (False, True),
        (0, False),
        ("", False),
    ],
)
def test_is_boolean(case, expected):
    assert _.is_boolean(case) == expected


@parametrize(
    "case,expected",
    [
        (list, True),
        (dict, True),
        ([].append, True),
        ({}.update, True),
        (1, True),
        (True, True),
        (None, True),
        (datetime, False),
        (parametrize, False),
        ("a", False),
        (lambda: None, False),
    ],
)
def test_is_builtin(case, expected):
    assert _.is_builtin(case) == expected


@parametrize(
    "case,expected",
    [
        (datetime.date.today(), True),
        (datetime.datetime.today(), True),
        ("2014-01-01", False),
        ("2014-01-01T00:00:00", False),
        (1, False),
    ],
)
def test_is_date(case, expected):
    assert _.is_date(case) == expected


@parametrize(
    "case,expected",
    [
        ([3, 2, 1], True),
        ([6, 5, 5], True),
        (0, True),
        ([5, 4, 4, 3, 1], True),
        ([5, 4, 4, 5, 4, 3], False),
    ],
)
def test_is_decreasing(case, expected):
    assert _.is_decreasing(case) == expected


@parametrize(
    "case,expected",
    [({}, True), ([], False), (1, False), ("a", False), (iter([]), False), (iter({}), False)],
)
def test_is_dict(case, expected):
    assert _.is_dict(case) == expected


@parametrize(
    "case,expected",
    [
        (True, True),
        (0, True),
        (123.45, True),
        ("", True),
        ({}, True),
        ([], True),
        (False, True),
        (None, True),
        ({"a": 1}, False),
        ([1], False),
        ("Hello", False),
        (["Hello", "World"], False),
    ],
)
def test_is_empty(case, expected):
    assert _.is_empty(case) == expected


@parametrize(
    "case,expected",
    [
        ((1, 1), True),
        ((1, 2), False),
        (("1", "1"), True),
        (("1", "2"), False),
        (([1], {"a": 1}), False),
        (({"a": 1}, {"a": 1}), True),
        (({"a": 1}, {"b": 1}), False),
        (([1, 2, 3], [1, 2, 3]), True),
        (([1, 2, 3], [1, 2]), False),
        (([1, 2], [1, 2, 3]), False),
        (([1, 2, 3], [1, 2]), False),
        (([1, 2], [1, 2, 3]), False),
        ((["hello", "goodbye"], ["hi", "goodbye"]), False),
    ],
)
def test_is_equal(case, expected):
    assert _.is_equal(*case) == expected


@parametrize(
    "case,expected",
    [
        ((1, 1, None), True),
        ((1, 2, None), False),
        (("1", "1", None), True),
        (("1", "2", None), False),
        (([1], {"a": 1}, None), False),
        (([1], {"a": 1}, lambda a, b: True), True),
        (({"a": 1}, {"a": 1}, None), True),
        (({"a": 1}, {"b": 1}, lambda a, b: None if isinstance(a, dict) else True), False),
        (([1, 2, 3], [1, 2, 3], None), True),
        (([1, 2, 3], [1, 2], None), False),
        (([1, 2], [1, 2, 3], None), False),
        (([1, 2, 3], [1, 2], lambda a, b: None if isinstance(a, list) else True), False),
        (([1, 2], [1, 2, 3], lambda a, b: None if isinstance(a, list) else True), False),
        ((["hello", "goodbye"], ["hi", "goodbye"], helpers.is_equal_iteratee0), True),
    ],
)
def test_is_equal_with(case, expected):
    assert _.is_equal_with(*case) == expected


@parametrize("case,expected", [(Exception(), True), ({}, False), ([], False)])
def test_is_error(case, expected):
    assert _.is_error(case) == expected


@parametrize(
    "case,expected",
    [
        (2, True),
        (16, True),
        (1, False),
        (3.0, False),
        (3.5, False),
        (None, False),
    ],
)
def test_is_even(case, expected):
    assert _.is_even(case) == expected


@parametrize(
    "case,expected",
    [
        (1.0, True),
        (3.245, True),
        (1, False),
        (True, False),
        ("", False),
    ],
)
def test_is_float(case, expected):
    assert _.is_float(case) == expected


@parametrize(
    "case,expected",
    [
        (lambda x: x + 1, True),
        ("Hello, world!", False),
    ],
)
def test_is_function(case, expected):
    assert _.is_function(case) == expected


@parametrize(
    "case,expected",
    [
        ([1, 2, 3], True),
        ([5, 5, 6], True),
        (0, True),
        ([1, 2, 3, 4, 4, 5], True),
        ([1, 2, 3, 4, 4, 3], False),
    ],
)
def test_is_increasing(case, expected):
    assert _.is_increasing(case) == expected


@parametrize(
    "case,expected",
    [
        ([], True),
        ("", True),
        ((), True),
        ({}, False),
        (1, False),
        (None, False),
    ],
)
def test_is_indexed(case, expected):
    assert _.is_indexed(case) == expected


@parametrize(
    "case,expected",
    [
        ((1, int), True),
        ((1.0, float), True),
        (([], (list, str)), True),
        (([], dict), False),
        ((True, float), False),
    ],
)
def test_is_instance_of(case, expected):
    assert _.is_instance_of(*case) == expected


@parametrize(
    "case,expected",
    [
        (1, True),
        (2, True),
        (1.0, False),
        (True, False),
        (None, False),
        ([], False),
    ],
)
def test_is_integer(case, expected):
    assert _.is_integer(case) == expected


@parametrize(
    "case,expected",
    [
        ([], True),
        ({}, True),
        ((), True),
        ("a", True),
        (5, False),
        (None, False),
    ],
)
def test_is_iterable(case, expected):
    assert _.is_iterable(case) == expected


@parametrize(
    "case,expected",
    [
        ('{"one": 1, "two": {"three": "3"}, "four": [4]}', True),
        ({"one": 1, "two": {"three": "3"}, "four": [4]}, False),
        ("", False),
        (1, False),
        (True, False),
    ],
)
def test_is_json(case, expected):
    assert _.is_json(case) == expected


@parametrize(
    "case,expected",
    [
        ([1, 2, 3], True),
        ({}, False),
    ],
)
def test_is_list(case, expected):
    assert _.is_list(case) == expected


@parametrize(
    "case,expected",
    [
        (({"name": "fred", "age": 40}, {"age": 40}), True),
        (({"name": "fred", "age": 40}, {"age": 40, "active": True}), False),
        (([1, 2, 3], [1, 2]), True),
        (([1, 2, 3], [1, 2, 3, 4]), False),
        (({}, {}), True),
        (({"a": 1}, {}), True),
        (([], []), True),
        (([1], []), True),
    ],
)
def test_is_match(case, expected):
    assert _.is_match(*case) == expected


@parametrize(
    "case,expected",
    [
        (([1, 2], [2, 4], lambda a, b: None if isinstance(a, list) else b == a + a), True),
        (([1, 2], [2, 4], lambda a, b, key: a == b == key), False),
        (([0, 1], [0, 1], lambda a, b, key: a == b == key), True),
    ],
)
def test_is_match_with(case, expected):
    assert _.is_match_with(*case) == expected


@parametrize(
    "case,expected",
    [
        (([1, 2, 3], operator.le), True),
        (([3, 2, 1], operator.ge), True),
        (([1, 1, 2], operator.lt), False),
        (([3, 3, 2], operator.gt), False),
    ],
)
def test_is_monotone(case, expected):
    assert _.is_monotone(*case) == expected


@parametrize(
    "case,expected",
    [
        (0, False),
        (123456789123456789123456789, False),
        (123.45, False),
        (decimal.Decimal(1), False),
        ("1", True),
    ],
)
def test_is_nan(case, expected):
    assert _.is_nan(case) == expected


@parametrize(
    "case,expected",
    [
        (-1, True),
        (-1.25, True),
        (-0.1, True),
        (0, False),
        (1, False),
        (True, False),
        (False, False),
    ],
)
def test_is_negative(case, expected):
    assert _.is_negative(case) == expected


@parametrize(
    "case,expected",
    [
        (None, True),
        (0, False),
    ],
)
def test_is_none(case, expected):
    assert _.is_none(case) == expected


@parametrize(
    "case,expected",
    [
        (0, True),
        (123456789123456789123456789, True),
        (123.45, True),
        (decimal.Decimal(1), True),
        ("1", False),
        (True, False),
        (False, False),
    ],
)
def test_is_number(case, expected):
    assert _.is_number(case) == expected


@parametrize(
    "case,expected",
    [({}, True), ([], True), (1, False), ("a", False), (iter([]), False), (iter({}), False)],
)
def test_is_object(case, expected):
    assert _.is_object(case) == expected


@parametrize(
    "case,expected",
    [
        (1, True),
        (3.0, True),
        (3.5, True),
        (2, False),
        (16, False),
        (None, False),
    ],
)
def test_is_odd(case, expected):
    assert _.is_odd(case) == expected


@parametrize(
    "case,expected",
    [(1, True), (1.25, True), (0.1, True), (-1, False), (0, False), (True, False), (False, False)],
)
def test_is_positive(case, expected):
    assert _.is_positive(case) == expected


@parametrize(
    "case,expected",
    [
        (re.compile(""), True),
        ("", False),
        ("Hello, world!", False),
        (1, False),
        ({}, False),
        ([], False),
        (None, False),
    ],
)
def test_is_reg_exp(case, expected):
    assert _.is_reg_exp(case) == expected


@parametrize("case,expected", [(set(), True), ([1, 2, 3], False)])
def test_is_set(case, expected):
    assert _.is_set(case) == expected


@parametrize(
    "case,expected",
    [
        ([1, 2, 3], False),
        ([3, 2, 1], True),
        ([1, 1, 2], False),
        ([3, 3, 2], False),
    ],
)
def test_is_strictly_decreasing(case, expected):
    assert _.is_strictly_decreasing(case) == expected


@parametrize(
    "case,expected",
    [
        ([1, 2, 3], True),
        ([3, 2, 1], False),
        ([1, 1, 2], False),
        ([3, 3, 2], False),
    ],
)
def test_is_strictly_increasing(case, expected):
    assert _.is_strictly_increasing(case) == expected


@parametrize(
    "case,expected",
    [("", True), ("Hello, world!", True), (1, False), ({}, False), ([], False), (None, False)],
)
def test_is_string(case, expected):
    assert _.is_string(case) == expected


@parametrize(
    "case,expected",
    [
        ((), True),
        ([], False),
        ({}, False),
    ],
)
def test_is_tuple(case, expected):
    assert _.is_tuple(case) == expected


@parametrize(
    "case,expected",
    [
        (0, True),
        (0.0, False),
        ("", False),
        (True, False),
        (False, False),
    ],
)
def test_is_zero(case, expected):
    assert _.is_zero(case) == expected



class Object(object):
    def __init__(self, **attrs):
        for key, value in attrs.items():
            setattr(self, key, value)


class ItemsObject(object):
    def __init__(self, items):
        self._items = items

    def items(self):
        if isinstance(self._items, dict):
            return list(self._items.items())
        else:
            return enumerate(self._items)


class IteritemsObject(object):
    def __init__(self, items):
        self._items = items

    def iteritems(self):
        if isinstance(self._items, dict):
            for key, value in self._items.items():
                yield key, value
        else:
            for i, item in enumerate(self._items):
                yield i, item


class Filter(object):
    def __init__(self, predicate):
        self.predicate = predicate

    def __call__(self, item):
        return self.predicate(item)


def reduce_iteratee0(total, num):
    return total + num


def reduce_iteratee1(result, num, key):
    result[key] = num * 3
    return result


def reduce_right_iteratee0(a, b):
    return a + b


def noop(*args, **kwargs):
    pass


def transform_iteratee0(result, num):
    num *= num
    if num % 2:
        result.append(num)
        return len(result) < 3


def is_equal_iteratee0(a, b):
    a_greet = a.startswith("h") if hasattr(a, "startswith") else False
    b_greet = b.startswith("h") if hasattr(b, "startswith") else False

    return a_greet == b_greet if a_greet or b_greet else None


def for_in_iteratee0(value, key, obj):
    obj[key] += value


def for_in_iteratee1(value, key, obj):
    obj[key] += value
    return False


def for_in_iteratee2(value, index, obj):
    if index == 2:
        obj[index] = "index:2"
        return True
    elif index == 0:
        obj[index] = False
        return True
    else:
        obj[index] = True
        return False



import time
from unittest import mock

import pytest

import pydash as _


parametrize = pytest.mark.parametrize


@parametrize("case,expected", [((lambda a, b: a / b, 4, 2), 2)])
def test_attempt(case, expected):
    assert _.attempt(*case) == expected


@parametrize("case,expected", [((lambda a, b: a / b, 4, 0), ZeroDivisionError)])
def test_attempt_exception(case, expected):
    assert isinstance(_.attempt(*case), expected)


@parametrize(
    "pairs,case,expected",
    [
        (
            (
                [_.matches({"b": 2}), _.constant("matches B")],
                [_.matches({"a": 1}), _.constant("matches A")],
            ),
            {"a": 1, "b": 2},
            "matches B",
        ),
        (
            ([_.matches({"b": 2}), _.constant("matches B")], [_.matches({"a": 1}), _.invert]),
            {"a": 1, "b": 3},
            {1: "a", 3: "b"},
        ),
        (
            (
                [_.matches({"a": 1}), _.constant("matches A")],
                [_.matches({"b": 2}), _.constant("matches B")],
            ),
            {"a": 1, "b": 2},
            "matches A",
        ),
    ],
)
def test_cond(pairs, case, expected):
    func = _.cond(*pairs)
    assert func(case) == expected


@parametrize(
    "case,expected",
    [
        ([_.matches({"b": 2})], ValueError),
        ([_.matches({"b": 2}), _.matches({"a": 1}), _.constant("matches B")], ValueError),
        ([1, 2], ValueError),
        ([[1, 2]], TypeError),
        ([_.matches({"b": 2}), 2], ValueError),
    ],
)
def test_cond_exception(case, expected):
    with pytest.raises(expected):
        _.cond(case)


@parametrize(
    "source,case,expected",
    [
        ({"a": lambda n: n == 0, "b": lambda n: n < 0}, {"a": 0, "b": -1}, True),
        ({"a": lambda n: n == 0, "b": lambda n: n < 0}, {"a": 1, "b": -1}, False),
        ({"a": lambda n: n == 0, "b": lambda n: n < 0}, {"a": 0, "b": 1}, False),
        ({"b": lambda n: n > 1}, {"b": 2}, True),
        ({"b": lambda n: n > 1}, {"b": 0}, False),
        ([lambda n: n == 0, lambda n: n < 1], [0, -1], True),
        ([lambda n: n == 0, lambda n: n < 1], [1, -1], False),
        ([lambda n: n == 0, lambda n: n < 1], [0, 1], False),
    ],
)
def test_conforms(source, case, expected):
    func = _.conforms(source)
    assert func(case) == expected


@parametrize(
    "source,case,expected",
    [
        ({"a": lambda n: n == 0, "b": lambda n: n < 0}, {"a": 0, "b": -1}, True),
        ({"a": lambda n: n == 0, "b": lambda n: n < 0}, {"a": 1, "b": -1}, False),
        ({"a": lambda n: n == 0, "b": lambda n: n < 0}, {"a": 0, "b": 1}, False),
        ({"b": lambda n: n > 1}, {"b": 2}, True),
        ({"b": lambda n: n > 1}, {"b": 0}, False),
        ([lambda n: n == 0, lambda n: n < 1], [0, -1], True),
        ([lambda n: n == 0, lambda n: n < 1], [1, -1], False),
        ([lambda n: n == 0, lambda n: n < 1], [0, 1], False),
    ],
)
def test_conforms_to(source, case, expected):
    assert _.conforms_to(case, source) == expected


@parametrize("case", ["foo", "bar"])
def test_constant(case):
    assert _.constant(case)() == case


@parametrize(
    "case,expected",
    [
        (([1, 10, 20]), 1),
        (([None, 10, 20]), 10),
        (([None, None, 20]), 20),
        (([None, [1, 2], [3, 4]]), [1, 2]),
        (([None, None, [3, 4]]), [3, 4]),
        (([None, None, None]), None),
    ],
)
def test_default_to_any(case, expected):
    assert _.default_to_any(*case) == expected


@parametrize(
    "case,expected",
    [(([1, 10]), 1), (([None, 10]), 10), (([[1, 2], [3, 4]]), [1, 2]), (([None, [3, 4]]), [3, 4])],
)
def test_default_to(case, expected):
    assert _.default_to(*case) == expected


@parametrize("case,expected", [((1,), 1), ((1, 2), 1), ((), None)])
def test_identity(case, expected):
    assert _.identity(*case) == expected


@parametrize(
    "case,arg,expected",
    [
        ("name", [{"name": "fred", "age": 40}, {"name": "barney", "age": 36}], ["fred", "barney"]),
        (
            "friends.1",
            [
                {"name": "fred", "age": 40, "friends": ["barney", "betty"]},
                {"name": "barney", "age": 36, "friends": ["fred", "wilma"]},
            ],
            ["betty", "wilma"],
        ),
        (
            ["friends.1"],
            [
                {"name": "fred", "age": 40, "friends.1": "betty"},
                {"name": "barney", "age": 36, "friends.1": "wilma"},
            ],
            ["betty", "wilma"],
        ),
        (
            {"name": "fred"},
            [{"name": "fred", "age": 40}, {"name": "barney", "age": 36}],
            [True, False],
        ),
        (
            lambda obj: obj["age"],
            [{"name": "fred", "age": 40}, {"name": "barney", "age": 36}],
            [40, 36],
        ),
        (
            None,
            [{"name": "fred", "age": 40}, {"name": "barney", "age": 36}],
            [{"name": "fred", "age": 40}, {"name": "barney", "age": 36}],
        ),
        (
            ["name", "fred"],
            [{"name": "fred", "age": 40}, {"name": "barney", "age": 36}],
            [True, False],
        ),
        (1, [[0, 1], [2, 3], [4, 5]], [1, 3, 5]),
        (("a", "b.c", ["d", 0, "f"]), [{"a": 1, "b": {"c": 2}, "d": [{"f": 3}, 4]}], [[1, 2, 3]]),
        (("a",), [{"a": 1}], [[1]]),
    ],
)
def test_iteratee(case, arg, expected):
    getter = _.iteratee(case)
    assert _.map_(arg, getter) == expected


@parametrize(
    "case,arg,expected",
    [
        ({"age": 36}, {"name": "barney", "age": 36}, True),
        ({"age": 36}, {"name": "barney", "age": 40}, False),
        ({"a": {"b": 2}}, {"a": {"b": 2, "c": 3}}, True),
    ],
)
def test_matches(case, arg, expected):
    assert _.matches(case)(arg) is expected


@parametrize(
    "case,arg,expected",
    [
        (("a", 1), {"a": 1, "b": 2}, True),
        (("a", 2), {"a": 1, "b": 2}, False),
        ((1, 2), [1, 2, 3], True),
        ((1, 3), [1, 2, 3], False),
    ],
)
def test_matches_property(case, arg, expected):
    assert _.matches_property(*case)(arg) is expected


@parametrize(
    "case,args,kwargs,key",
    [
        ((lambda a, b: a + b,), (1, 2), {}, "(1, 2){}"),
        ((lambda a, b: a + b,), (1,), {"b": 2}, "(1,){'b': 2}"),
        ((lambda a, b: a + b, lambda a, b: a * b), (1, 2), {}, 2),
        ((lambda a, b: a + b, lambda a, b: a * b), (1,), {"b": 2}, 2),
    ],
)
def test_memoize(case, args, kwargs, key):
    memoized = _.memoize(*case)
    expected = case[0](*args, **kwargs)
    assert memoized(*args, **kwargs) == expected
    assert memoized.cache
    assert memoized.cache[key] == expected


@parametrize(
    "case,args,kwargs,expected",
    [
        (("a.b",), ({"a": {"b": lambda x, y: x + y}}, 1, 2), {}, 3),
        (
            ("a.b",),
            (
                {"a": {"b": lambda x, y: x + y}},
                1,
            ),
            {"y": 2},
            3,
        ),
        (
            ("a.b", 1),
            (
                {"a": {"b": lambda x, y: x + y}},
                2,
            ),
            {},
            3,
        ),
    ],
)
def test_method(case, args, kwargs, expected):
    assert _.method(*case)(*args, **kwargs) == expected


@parametrize(
    "case,args,kwargs,expected",
    [
        (({"a": {"b": lambda x, y: x + y}},), ("a.b", 1, 2), {}, 3),
        (
            ({"a": {"b": lambda x, y: x + y}},),
            (
                "a.b",
                1,
            ),
            {"y": 2},
            3,
        ),
    ],
)
def test_method_of(case, args, kwargs, expected):
    assert _.method_of(*case)(*args, **kwargs) == expected


@parametrize("case,expected", [((), None), ((1, 2, 3), None)])
def test_noop(case, expected):
    assert _.noop(*case) == expected


@parametrize(
    "args,pos,expected",
    [
        ([11, 22, 33, 44], 0, 11),
        ([11, 22, 33, 44], -1, 44),
        ([11, 22, 33, 44], -4, 11),
        ([11, 22, 33, 44], -5, None),
        ([11, 22, 33, 44], 4, None),
        ([11, 22, 33], "1", 22),
        ([11, 22, 33], "xyz", 11),
        ([11, 22, 33], 1.45, 33),
        ([11, 22, 33], 1.51, 33),
    ],
)
def test_nth_arg(args, pos, expected):
    func = _.nth_arg(pos)
    assert func(*args) == expected


def test_now():
    present = int(time.time() * 1000)
    # Add some leeway when comparing time.
    assert (present - 1) <= _.now() <= (present + 1)


@parametrize("funcs,data,expected", [([max, min], [1, 2, 3, 4], [4, 1])])
def test_over(funcs, data, expected):
    assert _.over(funcs)(*data) == expected


@parametrize(
    "funcs,data,expected",
    [([lambda x: x is not None, bool], [1], True), ([lambda x: x is None, bool], [1], False)],
)
def test_over_every(funcs, data, expected):
    assert _.over_every(funcs)(*data) == expected


@parametrize(
    "funcs,data,expected",
    [
        ([lambda x: x is not None, bool], [1], True),
        ([lambda x: x is None, bool], [1], True),
        ([lambda x: x is False, lambda y: y == 2], [True], False),
    ],
)
def test_over_some(funcs, data, expected):
    assert _.over_some(funcs)(*data) == expected


@parametrize(
    "case,arg,expected",
    [
        ("name", [{"name": "fred", "age": 40}, {"name": "barney", "age": 36}], ["fred", "barney"]),
        ("name", [{"name": "fred", "age": 40}, {"name": "barney", "age": 36}], ["fred", "barney"]),
        (
            "spouse.name",
            [
                {"name": "fred", "age": 40, "spouse": {"name": "wilma"}},
                {"name": "barney", "age": 36, "spouse": {"name": "betty"}},
            ],
            ["wilma", "betty"],
        ),
    ],
)
def test_property_(case, arg, expected):
    assert _.map_(arg, _.property_(case)) == expected


@parametrize(
    "case,arg,expected",
    [
        ({"name": "fred", "age": 40}, ["name", "age"], ["fred", 40]),
    ],
)
def test_property_of(case, arg, expected):
    assert _.map_(arg, _.property_of(case)) == expected


@parametrize("case,minimum,maximum", [((), 0, 1), ((25,), 0, 25), ((5, 10), 5, 10)])
def test_random(case, minimum, maximum):
    for _x in range(50):
        rnd = _.random(*case)
        assert isinstance(rnd, int)
        assert minimum <= rnd <= maximum


@parametrize(
    "case,floating,minimum,maximum",
    [
        ((), True, 0, 1),
        ((25,), True, 0, 25),
        ((5, 10), True, 5, 10),
        ((5.0, 10), False, 5, 10),
        ((5, 10.0), False, 5, 10),
        ((5.0, 10.0), False, 5, 10),
        ((5.0, 10.0), True, 5, 10),
    ],
)
def test_random_float(case, floating, minimum, maximum):
    for _x in range(50):
        rnd = _.random(*case, floating=floating)
        assert isinstance(rnd, float)
        assert minimum <= rnd <= maximum


@parametrize(
    "case,expected",
    [
        ((10,), [0, 1, 2, 3, 4, 5, 6, 7, 8, 9]),
        ((1, 11), [1, 2, 3, 4, 5, 6, 7, 8, 9, 10]),
        ((11, 1), [11, 10, 9, 8, 7, 6, 5, 4, 3, 2]),
        ((11, 1, -2), [11, 9, 7, 5, 3]),
        ((11, 1, 2), []),
        ((0, 30, 5), [0, 5, 10, 15, 20, 25]),
        ((0, -10, -1), [0, -1, -2, -3, -4, -5, -6, -7, -8, -9]),
        ((0,), []),
        ((), []),
    ],
)
def test_range_(case, expected):
    assert list(_.range_(*case)) == expected


@parametrize(
    "case,expected",
    [
        ((4,), [3, 2, 1, 0]),
        ((-4,), [-3, -2, -1, 0]),
        ((1, 5), [4, 3, 2, 1]),
        ((0, 20, 5), [15, 10, 5, 0]),
        ((0, -20, -5), [-15, -10, -5, -0]),
        ((0, -4, -1), [-3, -2, -1, 0]),
        ((1, 10), [9, 8, 7, 6, 5, 4, 3, 2, 1]),
        ((-1, 10), [9, 8, 7, 6, 5, 4, 3, 2, 1, 0, -1]),
        ((-1, -10), [-9, -8, -7, -6, -5, -4, -3, -2, -1]),
        ((1, 10, 2), [9, 7, 5, 3, 1]),
        ((1, 10, -2), []),
        ((-1, 10, 2), [9, 7, 5, 3, 1, -1]),
        ((-1, 10, -2), []),
        ((-1, -10, 2), []),
        ((-1, -10, -2), [-9, -7, -5, -3, -1]),
        ((10, 1), [2, 3, 4, 5, 6, 7, 8, 9, 10]),
        ((-10, 1), [0, -1, -2, -3, -4, -5, -6, -7, -8, -9, -10]),
        ((-10, -1), [-2, -3, -4, -5, -6, -7, -8, -9, -10]),
        ((10, 1, 2), []),
        ((10, 1, -2), [2, 4, 6, 8, 10]),
        ((-10, 1, 2), [0, -2, -4, -6, -8, -10]),
        ((-10, 1, -2), []),
        ((-10, -1, 2), [-2, -4, -6, -8, -10]),
        ((-10, -1, -2), []),
        ((1, 4, 0), [1, 1, 1]),
        ((0,), []),
        ((10,), [9, 8, 7, 6, 5, 4, 3, 2, 1, 0]),
        ((1, 5), [4, 3, 2, 1]),
        ((0, 20, 5), [15, 10, 5, 0]),
        ((0, -4, -1), [-3, -2, -1, 0]),
        ((0,), []),
        ((), []),
    ],
)
def test_range_right(case, expected):
    assert list(_.range_right(*case)) == expected


@parametrize(
    "case,expected",
    [
        (({"cheese": "crumpets", "stuff": lambda: "nonsense"}, "cheese"), "crumpets"),
        (({"cheese": "crumpets", "stuff": lambda: "nonsense"}, "stuff"), "nonsense"),
        (({"cheese": "crumpets", "stuff": lambda: "nonsense"}, "foo"), None),
        ((False, "foo"), None),
        ((False, "foo", "default"), "default"),
    ],
)
def test_result(case, expected):
    assert _.result(*case) == expected


@parametrize(
    "case,delay_count,delay_times",
    [
        ({}, 2, [0.5, 1.0]),
        ({"attempts": 1}, 0, []),
        ({"attempts": 3, "delay": 0.5, "scale": 2.0}, 2, [0.5, 1.0]),
        ({"attempts": 5, "delay": 1.5, "scale": 2.5}, 4, [1.5, 3.75, 9.375, 23.4375]),
        ({"attempts": 5, "delay": 1.5, "max_delay": 8.0, "scale": 2.5}, 4, [1.5, 3.75, 8.0, 8.0]),
    ],
)
def test_retry(mock_sleep, case, delay_count, delay_times):
    @_.retry(**case)
    def func():
        raise KeyError()

    with pytest.raises(KeyError):
        func()

    assert delay_count == mock_sleep.call_count

    delay_calls = [mock.call(time) for time in delay_times]
    assert delay_calls == mock_sleep.call_args_list


@parametrize(
    "case,delay_count",
    [({"attempts": 3}, 0), ({"attempts": 3}, 1), ({"attempts": 3}, 2), ({"attempts": 5}, 3)],
)
def test_retry_success(mock_sleep, case, delay_count):
    counter = {True: 0}

    @_.retry(**case)
    def func():
        if counter[True] != delay_count:
            counter[True] += 1
            raise Exception()
        return True

    result = func()

    assert result is True
    assert counter[True] == delay_count
    assert delay_count == mock_sleep.call_count


@parametrize(
    "case,unexpected_delay_times",
    [
        ({"jitter": 5, "delay": 2, "scale": 1, "attempts": 5}, [2, 2, 2, 2]),
        ({"jitter": 10, "delay": 3, "scale": 1.5, "attempts": 5}, [3, 4.5, 6.75, 10.125]),
        ({"jitter": 1.0, "delay": 3, "scale": 1.5, "attempts": 5}, [3, 4.5, 6.75, 10.125]),
    ],
)
def test_retry_jitter(mock_sleep, case, unexpected_delay_times):
    @_.retry(**case)
    def func():
        raise ValueError()

    with pytest.raises(ValueError):
        func()

    unexpected_delay_calls = [mock.call(time) for time in unexpected_delay_times]

    assert len(unexpected_delay_calls) == mock_sleep.call_count
    assert unexpected_delay_calls != mock_sleep.call_args_list


@parametrize(
    "case,raise_exc,delay_count",
    [
        ({"attempts": 1, "exceptions": (RuntimeError,)}, RuntimeError, 0),
        ({"attempts": 2, "exceptions": (RuntimeError,)}, RuntimeError, 1),
        ({"attempts": 2, "exceptions": (RuntimeError,)}, Exception, 0),
    ],
)
def test_retry_exceptions(mock_sleep, case, raise_exc, delay_count):
    @_.retry(**case)
    def func():
        raise raise_exc()

    with pytest.raises(raise_exc):
        func()

    assert delay_count == mock_sleep.call_count


def test_retry_on_exception(mock_sleep):
    attempts = 5
    error_count = {True: 0}

    def on_exception(exc):
        error_count[True] += 1

    @_.retry(attempts=attempts, on_exception=on_exception)
    def func():
        raise TypeError()

    with pytest.raises(TypeError):
        func()

    assert error_count[True] == attempts


@parametrize(
    "case,exception",
    [
        ({"attempts": 0}, ValueError),
        ({"attempts": "1"}, ValueError),
        ({"delay": -1}, ValueError),
        ({"delay": "1"}, ValueError),
        ({"max_delay": -1}, ValueError),
        ({"max_delay": "1"}, ValueError),
        ({"scale": 0}, ValueError),
        ({"scale": "1"}, ValueError),
        ({"jitter": -1}, ValueError),
        ({"jitter": "1"}, ValueError),
        ({"jitter": (1,)}, ValueError),
        ({"jitter": ("1", "2")}, ValueError),
        ({"exceptions": (1, 2)}, TypeError),
        ({"exceptions": 1}, TypeError),
        ({"exceptions": (Exception, 2)}, TypeError),
        ({"on_exception": 5}, TypeError),
    ],
)
def test_retry_invalid_args(case, exception):
    with pytest.raises(exception):
        _.retry(**case)


@parametrize("case,expected", [(_.times(2, _.stub_list), [[], []]), (_.stub_list(), [])])
def test_stub_list(case, expected):
    assert case == expected


@parametrize("case,expected", [(_.times(2, _.stub_dict), [{}, {}]), (_.stub_dict(), {})])
def test_stub_dict(case, expected):
    assert case == expected


@parametrize("case,expected", [(_.times(2, _.stub_false), [False, False]), (_.stub_false(), False)])
def test_stub_false(case, expected):
    assert case == expected


@parametrize("case,expected", [(_.times(2, _.stub_string), ["", ""]), (_.stub_string(), "")])
def test_stub_string(case, expected):
    assert case == expected


@parametrize("case,expected", [(_.times(2, _.stub_true), [True, True]), (_.stub_true(), True)])
def test_stub_true(case, expected):
    assert case == expected


@parametrize("case,expected", [((5, lambda i: i * i), [0, 1, 4, 9, 16]), ((5,), [0, 1, 2, 3, 4])])
def test_times(case, expected):
    assert _.times(*case) == expected


@parametrize(
    "case,expected",
    [
        ("a.b.c", ["a", "b", "c"]),
        ("a[0].b.c", ["a", 0, "b", "c"]),
        ("a[0][1][2].b.c", ["a", 0, 1, 2, "b", "c"]),
        ("a[0][1][2].b.c", ["a", 0, 1, 2, "b", "c"]),
        ("a[0][-1][-2].b.c", ["a", 0, -1, -2, "b", "c"]),
    ],
)
def test_to_path(case, expected):
    assert _.to_path(case) == expected


def test_unique_id_setup():
    _.utilities.ID_COUNTER = 0


@parametrize("case,expected", [((), "1"), (("foo",), "foo2")])
def test_unique_id(case, expected):
    assert _.unique_id(*case) == expected
